import numpy as np
import flax.linen as nn
import jax
import jax.numpy as jnp
import math
import optax
import os
import pickle

from ..utils.atomic_units import AtomicUnits as au  # CM1,THZ,BOHR,MPROT
from ..utils import Counter
from ..utils.deconvolution import (
    deconvolute_spectrum,
    kernel_lorentz_pot,
    kernel_lorentz,
)


def get_thermostat(simulation_parameters, dt, system_data, fprec, rng_key=None, restart_data={}):
    state = {}
    postprocess = None
    

    thermostat_name = str(simulation_parameters.get("thermostat", "LGV")).upper()
    compute_thermostat_energy = simulation_parameters.get(
        "include_thermostat_energy", False
    )

    kT = system_data.get("kT", None)
    nbeads = system_data.get("nbeads", None)
    mass = system_data["mass"]
    gamma = simulation_parameters.get("gamma", 1.0 / au.THZ) / au.FS
    species = system_data["species"]

    if nbeads is not None:
        trpmd_lambda = simulation_parameters.get("trpmd_lambda", 1.0)
        gamma = np.maximum(trpmd_lambda * system_data["omk"], gamma)

    if thermostat_name in ["LGV", "LANGEVIN", "FFLGV"]:
        assert rng_key is not None, "rng_key must be provided for QTB thermostat"
        assert kT is not None, "kT must be specified for QTB thermostat"
        assert gamma is not None, "gamma must be specified for QTB thermostat"
        rng_key, v_key = jax.random.split(rng_key)
        if nbeads is None:
            a1 = math.exp(-gamma * dt)
            a2 = jnp.asarray(((1 - a1 * a1) * kT / mass[:, None]) ** 0.5, dtype=fprec)
            vel = (
                jax.random.normal(v_key, (mass.shape[0], 3), dtype=fprec)
                * (kT / mass[:, None]) ** 0.5
            )
        else:
            if isinstance(gamma, float):
                gamma = np.array([gamma] * nbeads)
            assert isinstance(
                gamma, np.ndarray
            ), "gamma must be a float or a numpy array"
            assert gamma.shape[0] == nbeads, "gamma must have the same length as nbeads"
            a1 = np.exp(-gamma * dt)[:, None, None]
            a2 = jnp.asarray(
                ((1 - a1 * a1) * kT / mass[None, :, None]) ** 0.5, dtype=fprec
            )
            vel = (
                jax.random.normal(v_key, (nbeads, mass.shape[0], 3), dtype=fprec)
                * (kT / mass[:, None]) ** 0.5
            )

        state["rng_key"] = rng_key
        if compute_thermostat_energy:
            state["thermostat_energy"] = 0.0
        if thermostat_name == "FFLGV":
            def thermostat(vel, state):
                rng_key, noise_key = jax.random.split(state["rng_key"])
                noise = jax.random.normal(noise_key, vel.shape, dtype=vel.dtype)
                norm_vel = jnp.linalg.norm(vel, axis=-1, keepdims=True)
                dirvel = vel / norm_vel
                if compute_thermostat_energy:
                    v2 = (vel**2).sum(axis=-1)
                vel = a1 * vel + a2 * noise
                new_norm_vel = jnp.linalg.norm(vel, axis=-1, keepdims=True)
                vel = dirvel * new_norm_vel
                new_state = {**state, "rng_key": rng_key}
                if compute_thermostat_energy:
                    v2new = (vel**2).sum(axis=-1)
                    new_state["thermostat_energy"] = (
                        state["thermostat_energy"] + 0.5 * (mass * (v2 - v2new)).sum()
                    )

                return vel, new_state

        else:
            def thermostat(vel, state):
                rng_key, noise_key = jax.random.split(state["rng_key"])
                noise = jax.random.normal(noise_key, vel.shape, dtype=vel.dtype)
                if compute_thermostat_energy:
                    v2 = (vel**2).sum(axis=-1)
                vel = a1 * vel + a2 * noise
                new_state = {**state, "rng_key": rng_key}
                if compute_thermostat_energy:
                    v2new = (vel**2).sum(axis=-1)
                    new_state["thermostat_energy"] = (
                        state["thermostat_energy"] + 0.5 * (mass * (v2 - v2new)).sum()
                    )
                return vel, new_state

    elif thermostat_name in ["BUSSI"]:
        assert rng_key is not None, "rng_key must be provided for QTB thermostat"
        assert kT is not None, "kT must be specified for QTB thermostat"
        assert gamma is not None, "gamma must be specified for QTB thermostat"
        assert nbeads is None, "Bussi thermostat is not compatible with PIMD"
        rng_key, v_key = jax.random.split(rng_key)

        a1 = math.exp(-gamma * dt)
        a2 = (1 - a1) * kT
        vel = (
            jax.random.normal(v_key, (mass.shape[0], 3), dtype=fprec)
            * (kT / mass[:, None]) ** 0.5
        )

        state["rng_key"] = rng_key
        if compute_thermostat_energy:
            state["thermostat_energy"] = 0.0

        def thermostat(vel, state):
            rng_key, noise_key = jax.random.split(state["rng_key"])
            new_state = {**state, "rng_key": rng_key}
            noise = jax.random.normal(noise_key, vel.shape, dtype=vel.dtype)
            R2 = jnp.sum(noise**2)
            R1 = noise[0, 0]
            c = a2 / (mass[:, None] * vel**2).sum()
            d = (a1 * c) ** 0.5
            scale = (a1 + c * R2 + 2 * d * R1) ** 0.5
            if compute_thermostat_energy:
                dek = 0.5 * (mass[:, None] * vel**2).sum() * (scale**2 - 1)
                new_state["thermostat_energy"] = state["thermostat_energy"] + dek
            return scale * vel, new_state

    elif thermostat_name in [
        "GD",
        "DESCENT",
        "GRADIENT_DESCENT",
        "MIN",
        "MINIMIZE",
    ]:
        assert nbeads is None, "Gradient descent is not compatible with PIMD"
        a1 = math.exp(-gamma * dt)

        if nbeads is None:
            vel = jnp.zeros((mass.shape[0], 3), dtype=fprec)
        else:
            vel = jnp.zeros((nbeads, mass.shape[0], 3), dtype=fprec)

        def thermostat(vel, state):
            return a1 * vel, state

    elif thermostat_name in ["NVE", "NONE"]:
        if nbeads is None:
            vel = (
                jax.random.normal(rng_key, (mass.shape[0], 3), dtype=fprec)
                * (kT / mass[:, None]) ** 0.5
            )
            kTsys = jnp.sum(mass[:, None] * vel**2) / (mass.shape[0] * 3)
            vel = vel * (kT / kTsys) ** 0.5
        else:
            vel = (
                jax.random.normal(rng_key, (nbeads, mass.shape[0], 3), dtype=fprec)
                * (kT / mass[None, :, None]) ** 0.5
            )
            kTsys = jnp.sum(mass[None, :, None] * vel**2, axis=(1, 2)) / (
                mass.shape[0] * 3
            )
            vel = vel * (kT / kTsys[:, None, None]) ** 0.5
        thermostat = lambda x, s: (x, s)

    elif thermostat_name in ["NOSE", "NOSEHOOVER", "NOSE_HOOVER"]:
        assert gamma is not None, "gamma must be specified for QTB thermostat"
        ndof = mass.shape[0] * 3
        nkT = ndof * kT
        nose_mass = nkT / gamma**2
        assert nbeads is None, "Nose-Hoover is not compatible with PIMD"
        state["nose_s"] = 0.0
        state["nose_v"] = 0.0
        if compute_thermostat_energy:
            state["thermostat_energy"] = 0.0
        print(
            "# WARNING: Nose-Hoover thermostat is not well tested yet. Energy conservation is not guaranteed."
        )
        vel = (
            jax.random.normal(rng_key, (mass.shape[0], 3), dtype=fprec)
            * (kT / mass[:, None]) ** 0.5
        )

        def thermostat(vel, state):
            nose_s = state["nose_s"]
            nose_v = state["nose_v"]
            kTsys = jnp.sum(mass[:, None] * vel**2)
            nose_v = nose_v + (0.5 * dt / nose_mass) * (kTsys - nkT)
            nose_s = nose_s + dt * nose_v
            vel = jnp.exp(-nose_v * dt) * vel
            kTsys = jnp.sum(mass[:, None] * vel**2)
            nose_v = nose_v + (0.5 * dt / nose_mass) * (kTsys - nkT)
            new_state = {**state, "nose_s": nose_s, "nose_v": nose_v}

            if compute_thermostat_energy:
                new_state["thermostat_energy"] = (
                    nkT * nose_s + (0.5 * nose_mass) * nose_v**2
                )
            return vel, new_state

    elif thermostat_name in ["QTB", "ADQTB"]:
        assert nbeads is None, "QTB is not compatible with PIMD"
        qtb_parameters = simulation_parameters.get("qtb", None)
        assert (
            qtb_parameters is not None
        ), "qtb_parameters must be provided for QTB thermostat"
        assert rng_key is not None, "rng_key must be provided for QTB thermostat"
        assert kT is not None, "kT must be specified for QTB thermostat"
        assert gamma is not None, "gamma must be specified for QTB thermostat"
        assert species is not None, "species must be provided for QTB thermostat"
        rng_key, v_key = jax.random.split(rng_key)
        vel = (
            jax.random.normal(v_key, (mass.shape[0], 3), dtype=fprec)
            * (kT / mass[:, None]) ** 0.5
        )

        thermostat, postprocess, qtb_state = initialize_qtb(
            qtb_parameters,
            system_data,
            fprec=fprec,
            dt=dt,
            mass=mass,
            gamma=gamma,
            kT=kT,
            species=species,
            rng_key=rng_key,
            adaptive=thermostat_name.startswith("AD"),
            compute_thermostat_energy=compute_thermostat_energy,
        )
        state = {**state, **qtb_state}

    elif thermostat_name in ["ANNEAL", "ANNEALING"]:
        assert rng_key is not None, "rng_key must be provided for QTB thermostat"
        assert kT is not None, "kT must be specified for QTB thermostat"
        assert gamma is not None, "gamma must be specified for QTB thermostat"
        assert nbeads is None, "ANNEAL is not compatible with PIMD"
        a1 = math.exp(-gamma * dt)
        a2 = jnp.asarray(((1 - a1 * a1) * kT / mass[:, None]) ** 0.5, dtype=fprec)

        anneal_parameters = simulation_parameters.get("annealing", {})
        init_factor = anneal_parameters.get("init_factor", 1.0 / 25.0)
        assert init_factor > 0.0, "init_factor must be positive"
        final_factor = anneal_parameters.get("final_factor", 1.0 / 10000.0)
        assert final_factor > 0.0, "final_factor must be positive"
        nsteps = simulation_parameters.get("nsteps")
        anneal_steps = anneal_parameters.get("anneal_steps", 1.0)
        assert (
            anneal_steps < 1.0 and anneal_steps > 0.0
        ), "warmup_steps must be between 0 and nsteps"
        pct_start = anneal_parameters.get("warmup_steps", 0.3)
        assert (
            pct_start < 1.0 and pct_start > 0.0
        ), "warmup_steps must be between 0 and nsteps"

        anneal_type = anneal_parameters.get("type", "cosine").lower()
        if anneal_type == "linear":
            schedule = optax.linear_onecycle_schedule(
                peak_value=1.0,
                div_factor=1.0 / init_factor,
                final_div_factor=1.0 / final_factor,
                transition_steps=int(anneal_steps * nsteps),
                pct_start=pct_start,
                pct_final=1.0,
            )
        elif anneal_type == "cosine_onecycle":
            schedule = optax.cosine_onecycle_schedule(
                peak_value=1.0,
                div_factor=1.0 / init_factor,
                final_div_factor=1.0 / final_factor,
                transition_steps=int(anneal_steps * nsteps),
                pct_start=pct_start,
            )
        else:
            raise ValueError(f"Unknown anneal_type {anneal_type}")

        state["rng_key"] = rng_key
        state["istep_anneal"] = 0

        rng_key, v_key = jax.random.split(rng_key)
        Tscale = schedule(0)
        print(f"# ANNEAL: initial temperature = {Tscale*kT*au.KELVIN:.3e} K")
        vel = (
            jax.random.normal(v_key, (mass.shape[0], 3), dtype=fprec)
            * (kT * Tscale / mass[:, None]) ** 0.5
        )

        def thermostat(vel, state):
            rng_key, noise_key = jax.random.split(state["rng_key"])
            noise = jax.random.normal(noise_key, vel.shape, dtype=vel.dtype)

            Tscale = schedule(state["istep_anneal"]) ** 0.5
            vel = a1 * vel + a2 * Tscale * noise
            return vel, {
                **state,
                "rng_key": rng_key,
                "istep_anneal": state["istep_anneal"] + 1,
            }

    else:
        raise ValueError(f"Unknown thermostat {thermostat_name}")

    return thermostat, postprocess, state, vel,thermostat_name


def initialize_qtb(
    qtb_parameters,
    system_data,
    fprec,
    dt,
    mass,
    gamma,
    kT,
    species,
    rng_key,
    adaptive,
    compute_thermostat_energy=False,
):
    state = {}
    post_state = {}
    verbose = qtb_parameters.get("verbose", False)
    if compute_thermostat_energy:
        state["thermostat_energy"] = 0.0

    mass = jnp.asarray(mass, dtype=fprec)

    nat = species.shape[0]
    # define type indices
    species_set = set(species)
    nspecies = len(species_set)
    idx = {sp: i for i, sp in enumerate(species_set)}
    type_idx = np.array([idx[sp] for sp in species], dtype=np.int32)

    n_of_type = np.zeros(nspecies, dtype=np.int32)
    for i in range(nspecies):
        n_of_type[i] = (type_idx == i).nonzero()[0].shape[0]
    n_of_type = jnp.asarray(n_of_type, dtype=fprec)
    mass_idx = jax.ops.segment_sum(mass, type_idx, nspecies) / n_of_type

    corr_kin = qtb_parameters.get("corr_kin", -1)
    do_corr_kin = corr_kin <= 0
    if do_corr_kin:
        corr_kin = 1.0
    state["corr_kin"] = corr_kin
    post_state["corr_kin_prev"] = corr_kin
    post_state["do_corr_kin"] = do_corr_kin
    post_state["isame_kin"] = 0

    # spectra parameters
    omegasmear = np.pi / dt / 100.0
    Tseg = qtb_parameters.get("tseg", 1.0 / au.PS) * au.FS
    nseg = int(Tseg / dt)
    Tseg = nseg * dt
    dom = 2 * np.pi / (3 * Tseg)
    omegacut = qtb_parameters.get("omegacut", 15000.0 / au.CM1) / au.FS
    nom = int(omegacut / dom)
    omega = dom * np.arange((3 * nseg) // 2 + 1)
    cutoff = jnp.asarray(
        1.0 / (1.0 + np.exp((omega - omegacut) / omegasmear)), dtype=fprec
    )
    assert (
        omegacut < omega[-1]
    ), f"omegacut must be smaller than {omega[-1]*au.CM1} CM-1"

    # initialize gammar
    assert (
        gamma < 0.5 * omegacut
    ), "gamma must be much smaller than omegacut (at most 0.5*omegacut)"
    gammar_min = qtb_parameters.get("gammar_min", 0.1)
    # post_state["gammar"] = jnp.asarray(np.ones((nspecies, nom)), dtype=fprec)
    gammar = np.ones((nspecies, nom), dtype=float)
    try:
        for i, sp in enumerate(species_set):
            if not os.path.exists(f"QTB_spectra_{sp}.out"): continue
            data = np.loadtxt(f"QTB_spectra_{sp}.out")
            gammar[i] = data[:, 4]/(gamma*au.FS*au.THZ)
            print(f"# Restored gammar for species {sp} from QTB_spectra_{sp}.out")
    except Exception as e:
        print(f"# Could not restore gammar for all species with exception {e}. Starting from scratch.")
        gammar[:,:] = 1.0
    post_state["gammar"] = jnp.asarray(gammar, dtype=fprec)

    # Ornstein-Uhlenbeck correction for colored noise
    a1 = np.exp(-gamma * dt)
    OUcorr = jnp.asarray(
        (1.0 - 2.0 * a1 * np.cos(omega * dt) + a1**2) / (dt**2 * (gamma**2 + omega**2)),
        dtype=fprec,
    )

    # hbar schedule
    classical_kernel = qtb_parameters.get("classical_kernel", False)
    hbar = qtb_parameters.get("hbar", 1.0) * au.FS
    u = 0.5 * hbar * np.abs(omega) / kT
    theta = kT * np.ones_like(omega)
    if hbar > 0:
        theta[1:] *= u[1:] / np.tanh(u[1:])
    theta = jnp.asarray(theta, dtype=fprec)

    noise_key, post_state["rng_key"] = jax.random.split(rng_key)
    del rng_key
    post_state["white_noise"] = jax.random.normal(
        noise_key, (3 * nseg, nat, 3), dtype=jnp.float32
    )

    startsave = qtb_parameters.get("startsave", 1)
    counter = Counter(nseg, startsave=startsave)
    state["istep"] = 0
    post_state["nadapt"] = 0
    post_state["nsample"] = 0

    write_spectra = qtb_parameters.get("write_spectra", True)
    do_compute_spectra = write_spectra or adaptive

    if do_compute_spectra:
        state["vel"] = jnp.zeros((nseg, nat, 3), dtype=fprec)

        post_state["dFDT"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["mCvv"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["Cvf"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["Cff"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["dFDT_avg"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["mCvv_avg"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["Cvfg_avg"] = jnp.zeros((nspecies, nom), dtype=fprec)
        post_state["Cff_avg"] = jnp.zeros((nspecies, nom), dtype=fprec)

    if not adaptive:
        update_gammar = lambda x: x
    else:
        # adaptation parameters
        skipseg = qtb_parameters.get("skipseg", 1)

        adaptation_method = (
            str(qtb_parameters.get("adaptation_method", "ADABELIEF")).upper().strip()
        )
        authorized_methods = ["SIMPLE", "RATIO", "ADABELIEF"]
        assert (
            adaptation_method in authorized_methods
        ), f"adaptation_method must be one of {authorized_methods}"
        if adaptation_method == "SIMPLE":
            agamma = qtb_parameters.get("agamma", 1.0e-3) / au.FS
            assert agamma > 0, "agamma must be positive"
            a1_ad = agamma * Tseg  #  * gamma
            print(f"# ADQTB SIMPLE: agamma = {agamma*au.FS:.3f}")

            def update_gammar(post_state):
                g = post_state["dFDT"]
                gammar = post_state["gammar"] - a1_ad * g
                gammar = jnp.maximum(gammar_min, gammar)
                return {**post_state, "gammar": gammar}

        elif adaptation_method == "RATIO":
            tau_ad = qtb_parameters.get("tau_ad", 5.0 / au.PS) * au.FS
            tau_s = qtb_parameters.get("tau_s", 10 * tau_ad) * au.FS
            assert tau_ad > 0, "tau_ad must be positive"
            print(
                f"# ADQTB RATIO: tau_ad = {tau_ad*1e-3:.2f} ps, tau_s = {tau_s*1e-3:.2f} ps"
            )
            b1 = np.exp(-Tseg / tau_ad)
            b2 = np.exp(-Tseg / tau_s)
            post_state["mCvv_m"] = jnp.zeros((nspecies, nom), dtype=fprec)
            post_state["Cvf_m"] = jnp.zeros((nspecies, nom), dtype=fprec)
            post_state["n_adabelief"] = 0

            def update_gammar(post_state):
                n_adabelief = post_state["n_adabelief"] + 1
                mCvv_m = post_state["mCvv_m"] * b1 + post_state["mCvv"] * (1.0 - b1)
                Cvf_m = post_state["Cvf_m"] * b2 + post_state["Cvf"] * (1.0 - b2)
                mCvv = mCvv_m / (1.0 - b1**n_adabelief)
                Cvf = Cvf_m / (1.0 - b2**n_adabelief)
                # g = Cvf/(mCvv+1.e-8)-post_state["gammar"]
                gammar = Cvf / (mCvv + 1.0e-8)
                gammar = jnp.maximum(gammar_min, gammar)
                return {
                    **post_state,
                    "gammar": gammar,
                    "mCvv_m": mCvv_m,
                    "Cvf_m": Cvf_m,
                    "n_adabelief": n_adabelief,
                }

        elif adaptation_method == "ADABELIEF":
            agamma = qtb_parameters.get("agamma", 0.1)
            tau_ad = qtb_parameters.get("tau_ad", 1.0 / au.PS) * au.FS
            tau_s = qtb_parameters.get("tau_s", 100 * tau_ad) * au.FS
            assert tau_ad > 0, "tau_ad must be positive"
            assert tau_s > 0, "tau_s must be positive"
            assert agamma > 0, "agamma must be positive"
            print(
                f"# ADQTB ADABELIEF: agamma = {agamma:.3f}, tau_ad = {tau_ad*1.e-3:.2f} ps, tau_s = {tau_s*1.e-3:.2f} ps"
            )

            a1_ad = agamma * gamma  # * Tseg #* gamma
            b1 = np.exp(-Tseg / tau_ad)
            b2 = np.exp(-Tseg / tau_s)
            post_state["dFDT_m"] = jnp.zeros((nspecies, nom), dtype=fprec)
            post_state["dFDT_s"] = jnp.zeros((nspecies, nom), dtype=fprec)
            post_state["n_adabelief"] = 0

            def update_gammar(post_state):
                n_adabelief = post_state["n_adabelief"] + 1
                dFDT = post_state["dFDT"]
                dFDT_m = post_state["dFDT_m"] * b1 + dFDT * (1.0 - b1)
                dFDT_s = (
                    post_state["dFDT_s"] * b2
                    + (dFDT - dFDT_m) ** 2 * (1.0 - b2)
                    + 1.0e-8
                )
                # bias correction
                mt = dFDT_m / (1.0 - b1**n_adabelief)
                st = dFDT_s / (1.0 - b2**n_adabelief)
                gammar = post_state["gammar"] - a1_ad * mt / (st**0.5 + 1.0e-8)
                gammar = jnp.maximum(gammar_min, gammar)
                return {
                    **post_state,
                    "gammar": gammar,
                    "dFDT_m": dFDT_m,
                    "n_adabelief": n_adabelief,
                    "dFDT_s": dFDT_s,
                }
    
    #####################
    # RESTART
    restart_file = system_data["name"]+".qtb.restart"
    if os.path.exists(restart_file):
        with open(restart_file, "rb") as f:
            data = pickle.load(f)
            state["corr_kin"] = data["corr_kin"]
            post_state["corr_kin_prev"] = data["corr_kin"]
            post_state["isame_kin"] = data["isame_kin"]
            post_state["do_corr_kin"] = data["do_corr_kin"]
            print(f"# Restored QTB state from {restart_file}")

    def write_qtb_restart(state, post_state):
        with open(restart_file, "wb") as f:
            pickle.dump(
                {
                    "corr_kin": state["corr_kin"],
                    "corr_kin_prev": post_state["corr_kin_prev"],
                    "isame_kin": post_state["isame_kin"],
                    "do_corr_kin": post_state["do_corr_kin"],
                },
                f,
            )
    ######################

    def compute_corr_pot(niter=20, verbose=False):
        if classical_kernel or hbar == 0:
            return np.ones(nom)

        s_0 = np.array((theta / kT * cutoff)[:nom])
        s_out, s_rec, _ = deconvolute_spectrum(
            s_0,
            omega[:nom],
            gamma,
            niter,
            kernel=kernel_lorentz_pot,
            trans=True,
            symmetrize=True,
            verbose=verbose,
        )
        corr_pot = 1.0 + (s_out - s_0) / s_0
        columns = np.column_stack(
            (omega[:nom] * au.CM1, corr_pot - 1.0, s_0, s_out, s_rec)
        )
        np.savetxt(
            "corr_pot.dat", columns, header="omega(cm-1) corr_pot s_0 s_out s_rec"
        )
        return corr_pot

    def compute_corr_kin(post_state, niter=7, verbose=False):
        if not post_state["do_corr_kin"]:
            return post_state["corr_kin_prev"], post_state
        if classical_kernel or hbar == 0:
            return 1.0, post_state

        K_D = post_state.get("K_D", None)
        mCvv = (post_state["mCvv_avg"][:, :nom] * n_of_type[:, None]).sum(axis=0) / nat
        s_0 = np.array(mCvv * kT / theta[:nom] / post_state["corr_pot"])
        s_out, s_rec, K_D = deconvolute_spectrum(
            s_0,
            omega[:nom],
            gamma,
            niter,
            kernel=kernel_lorentz,
            trans=False,
            symmetrize=True,
            verbose=verbose,
            K_D=K_D,
        )
        s_out = s_out * theta[:nom] / kT
        s_rec = s_rec * theta[:nom] / kT * post_state["corr_pot"]
        mCvvsum = mCvv.sum()
        rec_ratio = mCvvsum / s_rec.sum()
        if rec_ratio < 0.95 or rec_ratio > 1.05:
            print(
                f"# WARNING: reconvolution error {rec_ratio} is too high, corr_kin was not updated"
            )
            return post_state["corr_kin_prev"], post_state

        corr_kin = mCvvsum / s_out.sum()
        if np.abs(corr_kin - post_state["corr_kin_prev"]) < 1.0e-4:
            isame_kin = post_state["isame_kin"] + 1
        else:
            isame_kin = 0

        # print("# corr_kin: ", corr_kin)
        do_corr_kin = post_state["do_corr_kin"]
        if isame_kin > 10:
            print(
                "# INFO: corr_kin is converged (it did not change for 10 consecutive segments)"
            )
            do_corr_kin = False

        return corr_kin, {
            **post_state,
            "corr_kin_prev": corr_kin,
            "isame_kin": isame_kin,
            "do_corr_kin": do_corr_kin,
            "K_D": K_D,
        }

    @jax.jit
    def ff_kernel(post_state):
        if classical_kernel:
            kernel = cutoff * (2 * gamma * kT / dt)
        else:
            kernel = theta * cutoff * OUcorr * (2 * gamma / dt)
        gamma_ratio = jnp.concatenate(
            (
                post_state["gammar"].T * post_state["corr_pot"][:, None],
                jnp.ones(
                    (kernel.shape[0] - nom, nspecies), dtype=post_state["gammar"].dtype
                ),
            ),
            axis=0,
        )
        return kernel[:, None] * gamma_ratio * mass_idx[None, :]

    @jax.jit
    def refresh_force(post_state):
        rng_key, noise_key = jax.random.split(post_state["rng_key"])
        white_noise = jnp.concatenate(
            (
                post_state["white_noise"][nseg:],
                jax.random.normal(
                    noise_key, (nseg, nat, 3), dtype=post_state["white_noise"].dtype
                ),
            ),
            axis=0,
        )
        amplitude = ff_kernel(post_state) ** 0.5
        s = jnp.fft.rfft(white_noise, 3 * nseg, axis=0) * amplitude[:, type_idx, None]
        force = jnp.fft.irfft(s, 3 * nseg, axis=0)[nseg : 2 * nseg]
        return force, {**post_state, "rng_key": rng_key, "white_noise": white_noise}

    @jax.jit
    def compute_spectra(force, vel, post_state):
        sf = jnp.fft.rfft(force / gamma, 3 * nseg, axis=0, norm="ortho")
        sv = jnp.fft.rfft(vel, 3 * nseg, axis=0, norm="ortho")
        Cvv = jnp.sum(jnp.abs(sv[:nom]) ** 2, axis=-1).T
        Cff = jnp.sum(jnp.abs(sf[:nom]) ** 2, axis=-1).T
        Cvf = jnp.sum(jnp.real(sv[:nom] * jnp.conj(sf[:nom])), axis=-1).T

        mCvv = (
            (dt / 3.0)
            * jnp.zeros_like(post_state["mCvv"]).at[type_idx].add(Cvv)
            * mass_idx[:, None]
            / n_of_type[:, None]
        )
        Cvf = (
            (dt / 3.0)
            * jnp.zeros_like(post_state["Cvf"]).at[type_idx].add(Cvf)
            / n_of_type[:, None]
        )
        Cff = (
            (dt / 3.0)
            * jnp.zeros_like(post_state["Cff"]).at[type_idx].add(Cff)
            / n_of_type[:, None]
        )
        dFDT = mCvv * post_state["gammar"] - Cvf

        nsinv = 1.0 / post_state["nsample"]
        b1 = 1.0 - nsinv
        dFDT_avg = post_state["dFDT_avg"] * b1 + dFDT * nsinv
        mCvv_avg = post_state["mCvv_avg"] * b1 + mCvv * nsinv
        Cvfg_avg = post_state["Cvfg_avg"] * b1 + Cvf / post_state["gammar"] * nsinv
        Cff_avg = post_state["Cff_avg"] * b1 + Cff * nsinv

        return {
            **post_state,
            "mCvv": mCvv,
            "Cvf": Cvf,
            "Cff": Cff,
            "dFDT": dFDT,
            "dFDT_avg": dFDT_avg,
            "mCvv_avg": mCvv_avg,
            "Cvfg_avg": Cvfg_avg,
            "Cff_avg": Cff_avg,
        }

    def write_spectra_to_file(post_state):
        mCvv_avg = np.array(post_state["mCvv_avg"])
        Cvfg_avg = np.array(post_state["Cvfg_avg"])
        Cff_avg = np.array(post_state["Cff_avg"]) * 3.0 / dt / (gamma**2)
        dFDT_avg = np.array(post_state["dFDT_avg"])
        gammar = np.array(post_state["gammar"])
        Cff_theo = np.array(ff_kernel(post_state))[:nom].T
        for i, sp in enumerate(species_set):
            ff_scale = au.KELVIN / ((2 * gamma / dt) * mass_idx[i])
            columns = np.column_stack(
                (
                    omega[:nom] * (au.FS * au.CM1),
                    mCvv_avg[i],
                    Cvfg_avg[i],
                    dFDT_avg[i],
                    gammar[i] * gamma * (au.FS * au.THZ),
                    Cff_avg[i] * ff_scale,
                    Cff_theo[i] * ff_scale,
                )
            )
            np.savetxt(
                f"QTB_spectra_{sp}.out",
                columns,
                fmt="%12.6f",
                header="#omega mCvv Cvf dFDT gammar Cff",
            )
        if verbose:
            print("# QTB spectra written.")

    if compute_thermostat_energy:
        state["qtb_energy_flux"] = 0.0

    @jax.jit
    def thermostat(vel, state):
        istep = state["istep"]
        dvel = dt * state["force"][istep] / mass[:, None]
        new_vel = vel * a1 + dvel
        new_state = {**state, "istep": istep + 1}
        if do_compute_spectra:
            vel2 = state["vel"].at[istep].set(vel * a1**0.5 + 0.5 * dvel)
            new_state["vel"] = vel2
        if compute_thermostat_energy:
            dek = 0.5 * (mass[:, None] * (vel**2 - new_vel**2)).sum()
            ekcorr = (
                0.5
                * (mass[:, None] * new_vel**2).sum()
                * (1.0 - 1.0 / state.get("corr_kin", 1.0))
            )
            new_state["qtb_energy_flux"] = state["qtb_energy_flux"] + dek
            new_state["thermostat_energy"] = new_state["qtb_energy_flux"] + ekcorr
        return new_vel, new_state

    @jax.jit
    def postprocess_work(state, post_state):
        if do_compute_spectra:
            post_state = compute_spectra(state["force"], state["vel"], post_state)
        if adaptive:
            post_state = jax.lax.cond(
                post_state["nadapt"] > skipseg, update_gammar, lambda x: x, post_state
            )
        new_force, post_state = refresh_force(post_state)
        return {**state, "force": new_force}, post_state

    def postprocess(state, post_state):
        counter.increment()
        if not counter.is_reset_step:
            return state, post_state
        post_state["nadapt"] += 1
        post_state["nsample"] = max(post_state["nadapt"] - startsave + 1, 1)
        if verbose:
            print("# Refreshing QTB forces.")
        state, post_state = postprocess_work(state, post_state)
        state["corr_kin"], post_state = compute_corr_kin(post_state)
        state["istep"] = 0
        if write_spectra:
            write_spectra_to_file(post_state)
        write_qtb_restart(state, post_state)
        return state, post_state

    post_state["corr_pot"] = jnp.asarray(compute_corr_pot(), dtype=fprec)

    state["force"], post_state = refresh_force(post_state)
    return thermostat, (postprocess, post_state), state



import os
import yaml
import sys
import jax
import io
import time
import jax.numpy as jnp
import numpy as np
import optax
from collections import defaultdict
import json
from copy import deepcopy
from pathlib import Path
import argparse
try:
    import torch
except ImportError:
    raise ImportError(
        "PyTorch is required for training models. Install the CPU version following instructions at https://pytorch.org/get-started/locally/"
    )
import random
from flax import traverse_util
import json
import shutil
import pickle

from flax.core import freeze, unfreeze
from .io import (
    load_configuration,
    load_dataset,
    load_model,
    TeeLogger,
    copy_parameters,
)
from .utils import (
    get_loss_definition,
    get_train_step_function,
    get_validation_function,
    linear_schedule,
)
from .optimizers import get_optimizer, get_lr_schedule
from ..utils import deep_update, AtomicUnits as au
from ..utils.io import human_time_duration
from ..models.preprocessing import AtomPadding, check_input, convert_to_jax

def save_restart_checkpoint(output_directory,stage,training_state,preproc_state,metrics_state):
    with open(output_directory + "/restart_checkpoint.pkl", "wb") as f:
            pickle.dump({
                "stage": stage,
                "training": training_state,
                "preproc_state": preproc_state,
                "metrics_state": metrics_state,
            }, f)

def load_restart_checkpoint(output_directory):
    restart_checkpoint_file = output_directory + "/restart_checkpoint.pkl"
    if not os.path.exists(restart_checkpoint_file):
        raise FileNotFoundError(
            f"Training state file not found: {restart_checkpoint_file}"
        )
    with open(restart_checkpoint_file, "rb") as f:
        restart_checkpoint = pickle.load(f)
    return restart_checkpoint


def main():
    parser = argparse.ArgumentParser(prog="fennol_train")
    parser.add_argument("config_file", type=str)
    parser.add_argument("--model_file", type=str, default=None)
    args = parser.parse_args()
    config_file = args.config_file
    model_file = args.model_file

    os.environ["OMP_NUM_THREADS"] = "1"
    sys.stdout = io.TextIOWrapper(
        open(sys.stdout.fileno(), "wb", 0), write_through=True
    )

    restart_training = False
    if os.path.isdir(config_file):
        output_directory = Path(config_file).absolute().as_posix()
        restart_training = True
        while output_directory.endswith("/"):
            output_directory = output_directory[:-1]
        config_file = output_directory + "/config.yaml"
        backup_dir = output_directory + f"_backup_{time.strftime('%Y-%m-%d-%H-%M-%S')}"
        shutil.copytree(output_directory, backup_dir)
        print("Restarting training from", output_directory)

    parameters = load_configuration(config_file)

    ### Set the device
    device: str = parameters.get("device", "cpu").lower()
    if device == "cpu":
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
    elif device.startswith("cuda") or device.startswith("gpu"):
        if ":" in device:
            num = device.split(":")[-1]
            os.environ["CUDA_VISIBLE_DEVICES"] = num
        else:
            os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        device = "gpu"

    _device = jax.devices(device)[0]
    jax.config.update("jax_default_device", _device)

    if restart_training:
        restart_checkpoint = load_restart_checkpoint(output_directory)
    else:
        restart_checkpoint = None

    # output directory
    if not restart_training:
        output_directory = parameters.get("output_directory", None)
        if output_directory is not None:
            if "{now}" in output_directory:
                output_directory = output_directory.replace(
                    "{now}", time.strftime("%Y-%m-%d-%H-%M-%S")
                )
            output_directory = Path(output_directory).absolute()
            if not output_directory.exists():
                output_directory.mkdir(parents=True)
            print("Output directory:", output_directory)
        else:
            output_directory = "."

    output_directory = str(output_directory) + "/"

    # copy config_file to output directory
    # config_name = Path(config_file).name
    config_ext = Path(config_file).suffix
    with open(config_file) as f_in:
        config_data = f_in.read()
    with open(output_directory + "/config" + config_ext, "w") as f_out:
        f_out.write(config_data)

    # set log file
    log_file = "train.log"  # parameters.get("log_file", None)
    logger = TeeLogger(output_directory + log_file)
    logger.bind_stdout()

    # set matmul precision
    enable_x64 = parameters.get("double_precision", False)
    jax.config.update("jax_enable_x64", enable_x64)
    fprec = "float64" if enable_x64 else "float32"
    parameters["fprec"] = fprec
    if enable_x64:
        print("Double precision enabled.")

    matmul_precision = parameters.get("matmul_prec", "highest").lower()
    assert matmul_precision in [
        "default",
        "high",
        "highest",
    ], "matmul_prec must be one of 'default','high','highest'"
    jax.config.update("jax_default_matmul_precision", matmul_precision)

    # set random seed
    rng_seed = parameters.get("rng_seed", np.random.randint(0, 2**32 - 1))
    print(f"rng_seed: {rng_seed}")
    rng_key = jax.random.PRNGKey(rng_seed)
    torch.manual_seed(rng_seed)
    np.random.seed(rng_seed)
    random.seed(rng_seed)
    np_rng = np.random.Generator(np.random.PCG64(rng_seed))

    try:
        if "stages" in parameters["training"]:
            ## train in stages ##
            params = deepcopy(parameters)
            stages = params["training"].pop("stages")
            assert isinstance(stages, dict), "'stages' must be a dict with named stages"
            model_file_stage = model_file
            print_stages_params = params["training"].get("print_stages_params", False)
            for i, (stage, stage_params) in enumerate(stages.items()):
                rng_key, subkey = jax.random.split(rng_key)
                print("")
                print(f"### STAGE {i+1}: {stage} ###")

                ## remove end_event from previous stage ##
                if i > 0 and "end_event" in params["training"]:
                    params["training"].pop("end_event")

                ## incrementally update training parameters ##
                params = deep_update(params, {"training": stage_params})
                if model_file_stage is not None:
                    ## load model from previous stage ##
                    params["model_file"] = model_file_stage

                if restart_training and restart_checkpoint["stage"] != i + 1:
                    print(f"Skipping stage {i+1} (already completed)")
                    continue

                if print_stages_params:
                    print("stage parameters:")
                    print(json.dumps(params, indent=2, sort_keys=False))

                ## train stage ##
                _, model_file_stage = train(
                    (subkey, np_rng),
                    params,
                    stage=i + 1,
                    output_directory=output_directory,
                    restart_checkpoint=restart_checkpoint,
                )

                restart_training = False
                restart_checkpoint = None
        else:
            ## single training stage ##
            train(
                (rng_key, np_rng),
                parameters,
                model_file=model_file,
                output_directory=output_directory,
                restart_checkpoint=restart_checkpoint,
            )
    except KeyboardInterrupt:
        print("Training interrupted by user.")
    finally:
        if log_file is not None:
            logger.unbind_stdout()
            logger.close()


def train(
    rng,
    parameters,
    model_file=None,
    stage=None,
    output_directory=None,
    restart_checkpoint=None,
):
    if output_directory is None:
        output_directory = "./"
    elif output_directory == "":
        output_directory = "./"
    elif not output_directory.endswith("/"):
        output_directory += "/"
    stage_prefix = f"_stage_{stage}" if stage is not None else ""

    if isinstance(rng, tuple):
        rng_key, np_rng = rng
    else:
        rng_key = rng
        np_rng = np.random.Generator(np.random.PCG64(np.random.randint(0, 2**32 - 1)))

    ### GET TRAINING PARAMETERS ###
    training_parameters = parameters.get("training", {})

    #### LOAD MODEL ####
    if restart_checkpoint is not None:
        model_key = None
        model_file = output_directory + "latest_model.fnx"
    else:
        rng_key, model_key = jax.random.split(rng_key)
    model = load_model(parameters, model_file, rng_key=model_key)

    ### SAVE INITIAL MODEL ###
    save_initial_model = (
        restart_checkpoint is None 
        and (stage is None or stage == 0)
    )
    if save_initial_model:
        model.save(output_directory + "initial_model.fnx")

    model_ref = None
    if "model_ref" in training_parameters:
        model_ref = load_model(parameters, training_parameters["model_ref"])
        print("Reference model:", training_parameters["model_ref"])

        if "ref_parameters" in training_parameters:
            ref_parameters = training_parameters["ref_parameters"]
            assert isinstance(
                ref_parameters, list
            ), "ref_parameters must be a list of str"
            print("Reference parameters:", ref_parameters)
            model.variables = copy_parameters(
                model.variables, model_ref.variables, ref_parameters
            )

    ### SET FLOATING POINT PRECISION ###
    fprec = parameters.get("fprec", "float32")

    def convert_to_fprec(x):
        if jnp.issubdtype(x.dtype, jnp.floating):
            return x.astype(fprec)
        return x

    model.variables = jax.tree_map(convert_to_fprec, model.variables)

    ### SET UP LOSS FUNCTION ###
    loss_definition, used_keys, ref_keys = get_loss_definition(
        training_parameters, model_energy_unit=model.energy_unit
    )

    coordinates_ref_key = training_parameters.get("coordinates_ref_key", None)
    if coordinates_ref_key is not None:
        compute_ref_coords = True
        print("Reference coordinates:", coordinates_ref_key)
    else:
        compute_ref_coords = False

    #### LOAD DATASET ####
    dspath = training_parameters.get("dspath", None)
    if dspath is None:
        raise ValueError("Dataset path 'training/dspath' should be specified.")
    batch_size = training_parameters.get("batch_size", 16)
    batch_size_val = training_parameters.get("batch_size_val", None)
    rename_refs = training_parameters.get("rename_refs", {})
    training_iterator, validation_iterator = load_dataset(
        dspath=dspath,
        batch_size=batch_size,
        batch_size_val=batch_size_val,
        training_parameters=training_parameters,
        infinite_iterator=True,
        atom_padding=True,
        ref_keys=ref_keys,
        split_data_inputs=True,
        np_rng=np_rng,
        add_flags=["training"],
        fprec=fprec,
        rename_refs=rename_refs,
    )

    compute_forces = "forces" in used_keys
    compute_virial = "virial_tensor" in used_keys or "virial" in used_keys
    compute_stress = "stress_tensor" in used_keys or "stress" in used_keys
    compute_pressure = "pressure" in used_keys or "pressure_tensor" in used_keys

    ### get optimizer parameters ###
    max_epochs = int(training_parameters.get("max_epochs", 2000))
    epoch_format = len(str(max_epochs))
    nbatch_per_epoch = training_parameters.get("nbatch_per_epoch", 200)
    nbatch_per_validation = training_parameters.get("nbatch_per_validation", 20)

    schedule, sch_state, schedule_metrics,adaptive_scheduler = get_lr_schedule(max_epochs,nbatch_per_epoch,training_parameters)
    optimizer = get_optimizer(
        training_parameters, model.variables, schedule(sch_state)[0]
    )
    opt_st = optimizer.init(model.variables)

    ### exponential moving average of the parameters ###
    ema_decay = training_parameters.get("ema_decay", -1.0)
    if ema_decay > 0.0:
        assert ema_decay < 1.0, "ema_decay must be in (0,1)"
        ema = optax.ema(decay=ema_decay)
    else:
        ema = optax.identity()
    ema_st = ema.init(model.variables)

    #### end event ####
    end_event = training_parameters.get("end_event", None)
    if end_event is None or isinstance(end_event, str) and end_event.lower() == "none":
        is_end = lambda metrics: False
    else:
        assert len(end_event) == 2, "end_event must be a list of two elements"
        is_end = lambda metrics: metrics[end_event[0]] < end_event[1]

    if "energy_terms" in training_parameters:
        model.set_energy_terms(training_parameters["energy_terms"], jit=False)
    print("energy terms:", model.energy_terms)

    ### MODEL EVALUATION FUNCTION ###
    pbc_training = training_parameters.get("pbc_training", False)
    if compute_stress or compute_virial or compute_pressure:
        virial_key = "virial" if "virial" in used_keys else "virial_tensor"
        stress_key = "stress" if "stress" in used_keys else "stress_tensor"
        pressure_key = "pressure" if "pressure" in used_keys else "pressure_tensor"
        if compute_stress or compute_pressure:
            assert pbc_training, "PBC must be enabled for stress or virial training"
            print("Computing forces and stress tensor")
            def evaluate(model, variables, data):
                _, _, vir, output = model._energy_and_forces_and_virial(variables, data)
                cells = output["cells"]
                volume = jnp.abs(jnp.linalg.det(cells))
                stress = vir / volume[:, None, None]
                output[stress_key] = stress
                output[virial_key] = vir
                if pressure_key == "pressure":
                    output[pressure_key] = -jnp.trace(stress, axis1=1, axis2=2) / 3.0
                else:
                    output[pressure_key] = -stress
                return output

        else:
            print("Computing forces and virial tensor")
            def evaluate(model, variables, data):
                _, _, vir, output = model._energy_and_forces_and_virial(variables, data)
                output[virial_key] = vir
                return output

    elif compute_forces:
        print("Computing forces")
        def evaluate(model, variables, data):
            _, _, output = model._energy_and_forces(variables, data)
            return output

    elif model.energy_terms is not None:
        def evaluate(model, variables, data):
            _, output = model._total_energy(variables, data)
            return output

    else:
        def evaluate(model, variables, data):
            output = model.modules.apply(variables, data)
            return output

    ### TRAINING FUNCTIONS ###
    train_step = get_train_step_function(
        loss_definition=loss_definition,
        model=model,
        model_ref=model_ref,
        compute_ref_coords=compute_ref_coords,
        evaluate=evaluate,
        optimizer=optimizer,
        ema=ema,
    )

    validation = get_validation_function(
        loss_definition=loss_definition,
        model=model,
        model_ref=model_ref,
        compute_ref_coords=compute_ref_coords,
        evaluate=evaluate,
        return_targets=False,
    )

    #### CONFIGURE PREPROCESSING ####
    gpu_preprocessing = training_parameters.get("gpu_preprocessing", False)
    if gpu_preprocessing:
        print("GPU preprocessing activated.")

    minimum_image = training_parameters.get("minimum_image", False)
    preproc_state = unfreeze(model.preproc_state)
    layer_state = []
    for st in preproc_state["layers_state"]:
        stnew = unfreeze(st)
        #     st["nblist_skin"] = nblist_skin
        #     if nblist_stride > 1:
        #         st["skin_stride"] = nblist_stride
        #         st["skin_count"] = nblist_stride
        if pbc_training:
            stnew["minimum_image"] = minimum_image
        if "nblist_mult_size" in training_parameters:
            stnew["nblist_mult_size"] = training_parameters["nblist_mult_size"]
        if "nblist_add_neigh" in training_parameters:
            stnew["add_neigh"] = training_parameters["nblist_add_neigh"]
        if "nblist_add_atoms" in training_parameters:
            stnew["add_atoms"] = training_parameters["nblist_add_atoms"]
        layer_state.append(freeze(stnew))

    preproc_state["layers_state"] = tuple(layer_state)
    preproc_state["check_input"] = False
    model.preproc_state = freeze(preproc_state)

    ### INITIALIZE METRICS ###
    maes_prev = defaultdict(lambda: np.inf)
    metrics_beta = training_parameters.get("metrics_ema_decay", -1.0)
    smoothen_metrics = metrics_beta < 1.0 and metrics_beta > 0.0
    if smoothen_metrics:
        print("Computing smoothed metrics with beta =", metrics_beta)
        rmses_smooth = defaultdict(lambda: 0.0)
        maes_smooth = defaultdict(lambda: 0.0)
        rmse_tot_smooth = 0.0
        mae_tot_smooth = 0.0
        nsmooth = 0
    max_restore_count = training_parameters.get("max_restore_count", 5)    

    fmetrics = open(
        output_directory + f"metrics{stage_prefix}.traj",
        "w" if restart_checkpoint is None else "a",
    )

    keep_all_bests = training_parameters.get("keep_all_bests", False)
    save_model_at_epochs = training_parameters.get("save_model_at_epochs", [])
    save_model_at_epochs= set([int(x) for x in save_model_at_epochs])
    save_model_every = int(training_parameters.get("save_model_every_epoch", 0))
    if save_model_every > 0:
        save_model_every = list(range(save_model_every, max_epochs+save_model_every, save_model_every))
        save_model_at_epochs = save_model_at_epochs.union(save_model_every)
    save_model_at_epochs = list(save_model_at_epochs)
    save_model_at_epochs.sort()

    ### LR factor after restoring a model that has diverged
    restore_scaling = training_parameters.get("restore_scaling", 1)
    assert restore_scaling > 0.0, "restore_scaling must be > 0.0"
    assert restore_scaling <= 1.0, "restore_scaling must be <= 1.0"
    if restore_scaling != 1:
        print("Applying LR scaling after restore:", restore_scaling)


    metric_use_best = training_parameters.get("metric_best", "rmse_tot")  # .lower()
    metrics_state = {}

    ### INITIALIZE TRAINING STATE ###
    if restart_checkpoint is not None:
        training_state = deepcopy(restart_checkpoint["training"])
        epoch_start = training_state["epoch"]
        model.preproc_state = freeze(restart_checkpoint["preproc_state"])
        if smoothen_metrics:
            metrics_state = deepcopy(restart_checkpoint["metrics_state"])
            rmses_smooth = metrics_state["rmses_smooth"]
            maes_smooth = metrics_state["maes_smooth"]
            rmse_tot_smooth = metrics_state["rmse_tot_smooth"]
            mae_tot_smooth = metrics_state["mae_tot_smooth"]
            nsmooth = metrics_state["nsmooth"]
        print("Restored training state")
    else:
        epoch_start = 1
        training_state = {
            "rng_key": rng_key,
            "opt_state": opt_st,
            "ema_state": ema_st,
            "sch_state": sch_state,
            "variables": model.variables,
            "variables_ema": model.variables,
            "step": 0,
            "restore_count": 0,
            "restore_scale": 1,
            "epoch": 0,
            "best_metric": np.inf,
        }
        if smoothen_metrics:
            metrics_state = {
                "rmses_smooth": dict(rmses_smooth),
                "maes_smooth": dict(maes_smooth),
                "rmse_tot_smooth": rmse_tot_smooth,
                "mae_tot_smooth": mae_tot_smooth,
                "nsmooth": nsmooth,
            }
    
    del opt_st, ema_st, sch_state, preproc_state
    variables_save = training_state['variables']
    variables_ema_save = training_state['variables_ema']

    ### Training loop ###
    start = time.time()
    print("Starting training...")
    for epoch in range(epoch_start, max_epochs+1):
        training_state["epoch"] = epoch
        s = time.time()
        for _ in range(nbatch_per_epoch):
            # fetch data
            inputs0, data = next(training_iterator)

            # preprocess data
            inputs = model.preprocess(verbose=True,use_gpu=gpu_preprocessing,**inputs0)

            rng_key, subkey = jax.random.split(rng_key)
            inputs["rng_key"] = subkey
            inputs["training_epoch"] = epoch

            # adjust learning rate
            current_lr, training_state["sch_state"] = schedule(training_state["sch_state"])
            current_lr *= training_state["restore_scale"]
            training_state["opt_state"].inner_states["trainable"].inner_state[-1].hyperparams[
                "step_size"
            ] = current_lr 

            # train step
            loss,training_state, _ = train_step(data,inputs,training_state)

        rmses_avg = defaultdict(lambda: 0.0)
        maes_avg = defaultdict(lambda: 0.0)
        for _ in range(nbatch_per_validation):
            inputs0, data = next(validation_iterator)

            inputs = model.preprocess(verbose=True,use_gpu=gpu_preprocessing,**inputs0)

            rng_key, subkey = jax.random.split(rng_key)
            inputs["rng_key"] = subkey
            inputs["training_epoch"] = epoch

            rmses, maes, output_val = validation(
                data=data,
                inputs=inputs,
                variables=training_state["variables_ema"],
            )
            for k, v in rmses.items():
                rmses_avg[k] += v
            for k, v in maes.items():
                maes_avg[k] += v

        jax.block_until_ready(output_val)

        ### Timings ###
        e = time.time()
        epoch_time = e - s

        elapsed_time = e - start
        if not adaptive_scheduler:
            remain_glob = elapsed_time * (max_epochs - epoch + 1) / epoch
            remain_last = epoch_time * (max_epochs - epoch + 1)
            # estimate remaining time via weighted average (put more weight on last at the beginning)
            wremain = np.sin(0.5 * np.pi * epoch / max_epochs)
            remaining_time = human_time_duration(
                remain_glob * wremain + remain_last * (1 - wremain)
            )
        elapsed_time = human_time_duration(elapsed_time)
        batch_time = human_time_duration(
            epoch_time / (nbatch_per_epoch + nbatch_per_validation)
        )
        epoch_time = human_time_duration(epoch_time)

        for k in rmses_avg.keys():
            rmses_avg[k] /= nbatch_per_validation
        for k in maes_avg.keys():
            maes_avg[k] /= nbatch_per_validation

        ### Print metrics ###
        print("")
        line = f"Epoch {epoch}, lr={current_lr:.3e}, loss = {loss:.3e}"
        line += f", epoch time = {epoch_time}, batch time = {batch_time}"
        line += f", elapsed time = {elapsed_time}"
        if not adaptive_scheduler:
            line += f", est. remaining time = {remaining_time}"
        print(line)
        rmse_tot = 0.0
        mae_tot = 0.0
        for k in rmses_avg.keys():
            mult = loss_definition[k]["mult"]
            loss_prms = loss_definition[k]
            rmse_tot = rmse_tot + rmses_avg[k] * loss_prms["weight"] ** 0.5
            mae_tot = mae_tot + maes_avg[k] * loss_prms["weight"]
            unit = "(" + loss_prms["unit"] + ")" if "unit" in loss_prms else ""

            weight_str = ""
            if "weight_schedule" in loss_prms:
                w = linear_schedule(epoch, *loss_prms["weight_schedule"])
                weight_str = f"(w={w:.3f})"

            if rmses_avg[k] / mult < 1.0e-2:
                print(
                    f"    rmse_{k}= {rmses_avg[k]/mult:10.3e} ; mae_{k}= {maes_avg[k]/mult:10.3e}   {unit}  {weight_str}"
                )
            else:
                print(
                    f"    rmse_{k}= {rmses_avg[k]/mult:10.3f} ; mae_{k}= {maes_avg[k]/mult:10.3f}   {unit}  {weight_str}"
                )

        ### CHECK IF WE SHOULD RESTORE PREVIOUS MODEL ###
        restore = False
        reinit = False
        for k, mae in maes_avg.items():
            if np.isnan(mae):
                restore = True
                reinit = True
                print("NaN detected in mae")
                # for k, v in inputs.items():
                #     if hasattr(v,"shape"):
                #         if np.isnan(v).any():
                #             print(k,v)
                # sys.exit(1)
            if "threshold" in loss_definition[k]:
                thr = loss_definition[k]["threshold"]
                if mae > thr * maes_prev[k]:
                    restore = True
                    break

        epoch_time_str = time.strftime("%Y-%m-%d-%H-%M-%S")
        model.variables = training_state["variables_ema"]
        if epoch in save_model_at_epochs:
            filename = output_directory + f"model{stage_prefix}_epoch{epoch:0{epoch_format}}_{epoch_time_str}.fnx"
            print("Scheduled model save :", filename)
            model.save(filename)

        if restore:
            training_state["restore_count"] += 1
            if training_state["restore_count"]  > max_restore_count:
                if reinit:
                    raise ValueError("Model diverged and could not be restored.")
                else:
                    training_state["restore_count"]  = 0
                    print(
                        f"{max_restore_count} unsuccessful restores, resuming training."
                    )

            training_state["variables"] = deepcopy(variables_save)
            training_state["variables_ema"] = deepcopy(variables_ema_save)
            model.variables = training_state["variables_ema"]
            training_state["restore_scale"] *= restore_scaling
            print(f"Restored previous model after divergence. Restore scale: {training_state['restore_scale']:.3f}")
            if reinit:
                training_state["opt_state"] = optimizer.init(model.variables)
                training_state["ema_state"] = ema.init(model.variables)
                print("Reinitialized optimizer after divergence.")
            save_restart_checkpoint(output_directory,stage,training_state,model.preproc_state,metrics_state)
            continue

        training_state["restore_count"]  = 0

        variables_save = deepcopy(training_state["variables"])
        variables_ema_save = deepcopy(training_state["variables_ema"])
        maes_prev = maes_avg

        ### SAVE MODEL ###
        model.save(output_directory + "latest_model.fnx")

        # save metrics
        step = int(training_state["step"])
        metrics = {
            "epoch": epoch,
            "step": step,
            "data_count": step * batch_size,
            "elapsed_time": time.time() - start,
            "lr": current_lr,
            "loss": loss,
        }
        for k in rmses_avg.keys():
            mult = loss_definition[k]["mult"]
            metrics[f"rmse_{k}"] = rmses_avg[k] / mult
            metrics[f"mae_{k}"] = maes_avg[k] / mult

        metrics["rmse_tot"] = rmse_tot
        metrics["mae_tot"] = mae_tot
        if smoothen_metrics:
            nsmooth += 1
            for k in rmses_avg.keys():
                mult = loss_definition[k]["mult"]
                rmses_smooth[k] = (
                    metrics_beta * rmses_smooth[k] + (1.0 - metrics_beta) * rmses_avg[k]
                )
                maes_smooth[k] = (
                    metrics_beta * maes_smooth[k] + (1.0 - metrics_beta) * maes_avg[k]
                )
                metrics[f"rmse_smooth_{k}"] = (
                    rmses_smooth[k] / (1.0 - metrics_beta**nsmooth) / mult
                )
                metrics[f"mae_smooth_{k}"] = (
                    maes_smooth[k] / (1.0 - metrics_beta**nsmooth) / mult
                )
            rmse_tot_smooth = (
                metrics_beta * rmse_tot_smooth + (1.0 - metrics_beta) * rmse_tot
            )
            mae_tot_smooth = (
                metrics_beta * mae_tot_smooth + (1.0 - metrics_beta) * mae_tot
            )
            metrics["rmse_smooth_tot"] = rmse_tot_smooth / (1.0 - metrics_beta**nsmooth)
            metrics["mae_smooth_tot"] = mae_tot_smooth / (1.0 - metrics_beta**nsmooth)

            # update metrics state
            metrics_state["rmses_smooth"] = dict(rmses_smooth)
            metrics_state["maes_smooth"] = dict(maes_smooth)
            metrics_state["rmse_tot_smooth"] = rmse_tot_smooth
            metrics_state["mae_tot_smooth"] = mae_tot_smooth
            metrics_state["nsmooth"] = nsmooth

        assert (
            metric_use_best in metrics
        ), f"Error: metric for selectring best model '{metric_use_best}' not in metrics"

        metric_for_best = metrics[metric_use_best]
        if metric_for_best <  training_state["best_metric"]:
            training_state["best_metric"] = metric_for_best
            metrics["best_metric"] = training_state["best_metric"]
            if keep_all_bests:
                best_name = (
                    output_directory
                    + f"best_model{stage_prefix}_epoch{epoch:0{epoch_format}}_{epoch_time_str}.fnx"
                )
                model.save(best_name)

            best_name = output_directory + f"best_model{stage_prefix}.fnx"
            model.save(best_name)
            print("New best model saved to:", best_name)
        else:
            metrics["best_metric"] = training_state["best_metric"]

        if epoch == 1:
            headers = [f"{i+1}:{k}" for i, k in enumerate(metrics.keys())]
            fmetrics.write("# " + " ".join(headers) + "\n")
        fmetrics.write(" ".join([str(metrics[k]) for k in metrics.keys()]) + "\n")
        fmetrics.flush()

        # update learning rate using current metrics
        assert (
            schedule_metrics in metrics
        ), f"Error: cannot update lr, '{schedule_metrics}' not in metrics"
        current_lr, training_state["sch_state"] = schedule(training_state["sch_state"], metrics[schedule_metrics])

        # update and save training state
        save_restart_checkpoint(output_directory,stage,training_state,model.preproc_state,metrics_state)

        
        if is_end(metrics):
            print("Stage finished.")
            break

    end = time.time()

    print(f"Training time: {human_time_duration(end-start)}")
    print("")

    fmetrics.close()

    filename = output_directory + f"final_model{stage_prefix}.fnx"
    model.save(filename)
    print("Final model saved to:", filename)

    return metrics, filename


if __name__ == "__main__":
    main()



import flax.linen as nn
from typing import Any, Sequence, Callable, Union, ClassVar, Optional, Dict, List
import jax.numpy as jnp
import jax
import numpy as np
from functools import partial
from ...utils.activations import activation_from_str
from ...utils.periodic_table import (
    CHEMICAL_PROPERTIES,
    PERIODIC_TABLE,
    PERIODIC_TABLE_REV_IDX,
)


def apply_switch(x: jax.Array, switch: jax.Array):
    """@private Multiply a switch array to an array of values."""
    shape = x.shape
    return (
        jnp.expand_dims(x, axis=-1).reshape(*switch.shape, -1) * switch[..., None]
    ).reshape(shape)


class ApplySwitch(nn.Module):
    """Multiply an edge array by a switch array.

    FID: APPLY_SWITCH
    """

    key: str
    """The key of the input array."""
    switch_key: Optional[str] = None
    """The key of the switch array."""
    graph_key: Optional[str] = None
    """The key of the graph containing the switch."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""

    FID: ClassVar[str] = "APPLY_SWITCH"

    @nn.compact
    def __call__(self, inputs) -> Any:
        if self.graph_key is not None:
            graph = inputs[self.graph_key]
            switch = graph["switch"]
        elif self.switch_key is not None:
            switch = inputs[self.switch_key]
        else:
            raise ValueError("Either graph_key or switch_key must be specified")

        x = inputs[self.key]
        output = apply_switch(x, switch)
        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: output}


class AtomToEdge(nn.Module):
    """Map atom-wise values to edge-wise values.

    FID: ATOM_TO_EDGE

    By default, we map the destination atom value to the edge. This can be changed by setting `use_source` to True.
    """

    _graphs_properties: Dict
    key: str
    """The key of the input atom-wise array."""
    output_key: Optional[str] = None
    """The key of the output edge-wise array. If None, the input key is used."""
    graph_key: str = "graph"
    """The key of the graph containing the edges."""
    switch: bool = False
    """Whether to apply a switch to the edge values."""
    switch_key: Optional[str] = None
    """The key of the switch array. If None, the switch is taken from the graph."""
    use_source: bool = False
    """Whether to use the source atom value instead of the destination atom value."""

    FID: ClassVar[str] = "ATOM_TO_EDGE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        graph = inputs[self.graph_key]
        nat = inputs["species"].shape[0]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        x = inputs[self.key]
        if self.use_source:
            x_edge = x[edge_src]
        else:
            x_edge = x[edge_dst]

        if self.switch:
            switch = (
                graph["switch"] if self.switch_key is None else inputs[self.switch_key]
            )
            x_edge = apply_switch(x_edge, switch)

        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: x_edge}


class ScatterEdges(nn.Module):
    """Reduce an edge array to atoms by summing over neighbors.

    FID: SCATTER_EDGES
    """

    _graphs_properties: Dict
    key: str
    """The key of the input edge-wise array."""
    output_key: Optional[str] = None
    """The key of the output atom-wise array. If None, the input key is used."""
    graph_key: str = "graph"
    """The key of the graph containing the edges."""
    switch: bool = False
    """Whether to apply a switch to the edge values before summing."""
    switch_key: Optional[str] = None
    """The key of the switch array. If None, the switch is taken from the graph."""
    antisymmetric: bool = False

    FID: ClassVar[str] = "SCATTER_EDGES"

    @nn.compact
    def __call__(self, inputs) -> Any:
        graph = inputs[self.graph_key]
        nat = inputs["species"].shape[0]
        x = inputs[self.key]

        if self.switch:
            switch = (
                graph["switch"] if self.switch_key is None else inputs[self.switch_key]
            )
            x = apply_switch(x, switch)

        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        output = jax.ops.segment_sum(
            x, edge_src, nat
        )  # jnp.zeros((nat, *x.shape[1:])).at[edge_src].add(x,mode="drop")
        if self.antisymmetric:
            output = output - jax.ops.segment_sum(x, edge_dst, nat)

        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: output}


class EdgeConcatenate(nn.Module):
    """Concatenate the source and destination atom values of an edge.

    FID: EDGE_CONCATENATE
    """

    _graphs_properties: Dict
    key: str
    """The key of the input atom-wise array."""
    output_key: Optional[str] = None
    """The key of the output edge-wise array. If None, the input key is used."""
    graph_key: str = "graph"
    """The key of the graph containing the edges."""
    switch: bool = False
    """Whether to apply a switch to the edge values."""
    switch_key: Optional[str] = None
    """The key of the switch array. If None, the switch is taken from the graph."""
    axis: int = -1
    """The axis along which to concatenate the atom values."""

    FID: ClassVar[str] = "EDGE_CONCATENATE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        nat = inputs["species"].shape[0]
        xi = inputs[self.key]

        assert self._graphs_properties[self.graph_key][
            "directed"
        ], "EdgeConcatenate only works for directed graphs"
        assert xi.shape[0] == nat, "Shape mismatch, xi.shape[0] != nat"

        xij = jnp.concatenate([xi[edge_src], xi[edge_dst]], axis=self.axis)

        if self.switch:
            switch = (
                graph["switch"] if self.switch_key is None else inputs[self.switch_key]
            )
            xij = apply_switch(xij, switch)

        output_key = self.name if self.output_key is None else self.output_key
        return {**inputs, output_key: xij}


class ScatterSystem(nn.Module):
    """Reduce an atom-wise array to a system-wise array by summing over atoms (in the batch).

    FID: SCATTER_SYSTEM
    """

    key: str
    """The key of the input atom-wise array."""
    output_key: Optional[str] = None
    """The key of the output system-wise array. If None, the input key is used."""
    average: bool = False
    """Wether to divide by the number of atoms in the system."""

    FID: ClassVar[str] = "SCATTER_SYSTEM"

    @nn.compact
    def __call__(self, inputs) -> Any:
        batch_index = inputs["batch_index"]
        x = inputs[self.key]
        assert (
            x.shape[0] == batch_index.shape[0]
        ), f"Shape mismatch {x.shape[0]} != {batch_index.shape[0]}"
        nsys = inputs["natoms"].shape[0]
        if self.average:
            shape = [batch_index.shape[0]] + (x.ndim - 1) * [1]
            x = x / inputs["natoms"][batch_index].reshape(shape)

        output = jax.ops.segment_sum(x, batch_index, nsys)

        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: output}


class SystemToAtoms(nn.Module):
    """Broadcast a system-wise array to an atom-wise array.

    FID: SYSTEM_TO_ATOMS
    """

    key: str
    """The key of the input system-wise array."""
    output_key: Optional[str] = None
    """The key of the output atom-wise array. If None, the input key is used."""

    FID: ClassVar[str] = "SYSTEM_TO_ATOMS"

    @nn.compact
    def __call__(self, inputs) -> Any:
        batch_index = inputs["batch_index"]
        x = inputs[self.key]
        output = x[batch_index]

        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: output}


class SumAxis(nn.Module):
    """Sum an array along an axis.

    FID: SUM_AXIS
    """

    key: str
    """The key of the input array."""
    axis: Union[None, int, Sequence[int]] = None
    """The axis along which to sum the array."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""
    norm: Optional[str] = None
    """Normalization of the sum. Can be 'dim', 'sqrt', or 'none'."""

    FID: ClassVar[str] = "SUM_AXIS"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]
        output = jnp.sum(x, axis=self.axis)
        if self.norm is not None:
            norm = self.norm.lower()
            if norm == "dim":
                dim = np.prod(x.shape[self.axis])
                output = output / dim
            elif norm == "sqrt":
                dim = np.prod(x.shape[self.axis])
                output = output / dim**0.5
            elif norm == "none":
                pass
            else:
                raise ValueError(f"Unknown norm {norm}")
        output_key = self.key if self.output_key is None else self.output_key
        return {**inputs, output_key: output}


class Split(nn.Module):
    """Split an array along an axis.

    FID: SPLIT
    """

    key: str
    """The key of the input array."""
    output_keys: Sequence[str]
    """The keys of the output arrays."""
    axis: int = -1
    """The axis along which to split the array."""
    sizes: Union[int, Sequence[int]] = 1
    """The sizes of the splits."""
    squeeze: bool = True
    """Whether to remove the axis in the output if the size is 1."""

    FID: ClassVar[str] = "SPLIT"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]

        if isinstance(self.sizes, int):
            split_size = [self.sizes] * len(self.output_keys)
        else:
            split_size = self.sizes
        if len(split_size) == len(self.output_keys):
            assert (
                sum(split_size) == x.shape[self.axis]
            ), f"Split sizes {split_size} do not match input shape"
            split_size = split_size[:-1]
        assert (
            len(split_size) == len(self.output_keys) - 1
        ), f"Wrong number of split sizes {split_size} for {len(self.output_keys)} outputs"
        split_indices = np.cumsum(split_size)
        outs = {}

        for k, v in zip(self.output_keys, jnp.split(x, split_indices, axis=self.axis)):
            outs[k] = (
                jnp.squeeze(v, axis=self.axis)
                if self.squeeze and v.shape[self.axis] == 1
                else v
            )

        return {**inputs, **outs}


class Concatenate(nn.Module):
    """Concatenate a list of arrays along an axis.

    FID: CONCATENATE
    """

    keys: Sequence[str]
    """The keys of the input arrays."""
    axis: int = -1
    """The axis along which to concatenate the arrays."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the name of the module is used."""

    FID: ClassVar[str] = "CONCATENATE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        output = jnp.concatenate([inputs[k] for k in self.keys], axis=self.axis)
        output_key = self.output_key if self.output_key is not None else self.name
        return {**inputs, output_key: output}


class Activation(nn.Module):
    """Apply an element-wise activation function to an array.

    FID: ACTIVATION
    """

    key: str
    """The key of the input array."""
    activation: Union[Callable, str]
    """The activation function or its name."""
    scale_out: float = 1.0
    """Output scaling factor."""
    shift_out: float = 0.0
    """Output shift."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""

    FID: ClassVar[str] = "ACTIVATION"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]
        activation = (
            activation_from_str(self.activation)
            if isinstance(self.activation, str)
            else self.activation
        )
        output = self.scale_out * activation(x) + self.shift_out
        output_key = self.output_key if self.output_key is not None else self.key
        return {**inputs, output_key: output}


class Scale(nn.Module):
    """Scale an array by a constant factor.

    FID: SCALE
    """

    key: str
    """The key of the input array."""
    scale: float
    """The (initial) scaling factor."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""
    trainable: bool = False
    """Whether the scaling factor is trainable."""

    FID: ClassVar[str] = "SCALE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        x = inputs[self.key]

        if self.trainable:
            scale = self.param("scale", lambda rng: jnp.asarray(self.scale))
        else:
            scale = self.scale

        output = scale * x
        output_key = self.output_key if self.output_key is not None else self.key
        return {**inputs, output_key: output}


class Add(nn.Module):
    """Add together a list of arrays.

    FID: ADD
    """

    keys: Sequence[str]
    """The keys of the input arrays."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the name of the module is used."""

    FID: ClassVar[str] = "ADD"

    @nn.compact
    def __call__(self, inputs) -> Any:
        output = 0
        for k in self.keys:
            output = output + inputs[k]

        output_key = self.output_key if self.output_key is not None else self.name
        return {**inputs, output_key: output}


class Multiply(nn.Module):
    """Element-wise-multiply together a list of arrays.

    FID: MULTIPLY
    """

    keys: Sequence[str]
    """The keys of the input arrays."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the name of the module is used."""

    FID: ClassVar[str] = "MULTIPLY"

    @nn.compact
    def __call__(self, inputs) -> Any:
        output = 1
        for k in self.keys:
            output = output * inputs[k]

        output_key = self.output_key if self.output_key is not None else self.name
        return {**inputs, output_key: output}


class Transpose(nn.Module):
    """Transpose an array.

    FID: TRANSPOSE
    """

    key: str
    """The key of the input array."""
    axes: Sequence[int]
    """The permutation of the axes. See `jax.numpy.transpose` for more details."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""

    FID: ClassVar[str] = "TRANSPOSE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        output = jnp.transpose(inputs[self.key], axes=self.axes)
        output_key = self.output_key if self.output_key is not None else self.key
        return {**inputs, output_key: output}


class Reshape(nn.Module):
    """Reshape an array.

    FID: RESHAPE
    """

    key: str
    """The key of the input array."""
    shape: Sequence[Union[int,str]]
    """The shape of the output array."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the input key is used."""

    FID: ClassVar[str] = "RESHAPE"

    @nn.compact
    def __call__(self, inputs) -> Any:
        shape = []
        for s in self.shape:
            if isinstance(s,int):
                shape.append(s)
                continue

            if isinstance(s,str):
                s_=s.lower().strip()
                if s_ in ["natoms" ,"nat","natom","n_atoms","atoms"]:
                    shape.append(inputs["species"].shape[0])
                    continue
                
                if s_ in ["nsys","nbatch","nsystems","n_sys","n_systems","n_batch"]:
                    shape.append(inputs["natoms"].shape[0])
                    continue
                
                s_ = s.strip().split("[")
                key = s_[0]
                if key in inputs:
                    axis = int(s_[1].split("]")[0])
                    shape.append(inputs[key].shape[axis])
                    continue

            raise ValueError(f"Error parsing shape component {s}")

        output = jnp.reshape(inputs[self.key], shape)
        output_key = self.output_key if self.output_key is not None else self.key
        return {**inputs, output_key: output}


class ChemicalConstant(nn.Module):
    """Map atomic species to a constant value.

    FID: CHEMICAL_CONSTANT
    """

    value: Union[str, List[float], float, Dict]
    """The constant value or a dictionary of values for each element."""
    output_key: Optional[str] = None
    """The key of the output array. If None, the name of the module is used."""
    trainable: bool = False
    """Whether the constant is trainable."""

    FID: ClassVar[str] = "CHEMICAL_CONSTANT"

    @nn.compact
    def __call__(self, inputs) -> Any:
        if isinstance(self.value, str):
            constant = CHEMICAL_PROPERTIES[self.value.upper()]
        elif isinstance(self.value, list) or isinstance(self.value, tuple):
            constant = list(self.value)
        elif isinstance(self.value, float):
            constant = [self.value] * len(PERIODIC_TABLE)
        elif hasattr(self.value, "items"):
            constant = [0.0] * len(PERIODIC_TABLE)
            for k, v in self.value.items():
                constant[PERIODIC_TABLE_REV_IDX[k]] = v
        else:
            raise ValueError(f"Unknown constant type {type(self.value)}")

        if self.trainable:
            constant = self.param(
                "constant", lambda rng: jnp.asarray(constant, dtype=jnp.float32)
            )
        else:
            constant = jnp.asarray(constant, dtype=jnp.float32)
        output = constant[inputs["species"]]
        output_key = self.output_key if self.output_key is not None else self.name
        return {**inputs, output_key: output}


class SwitchFunction(nn.Module):
    """Compute a switch array from an array of distances and a cutoff.

    FID: SWITCH_FUNCTION
    """

    cutoff: Optional[float] = None
    """The cutoff distance. If None, the cutoff is taken from the graph."""
    switch_start: float = 0.0
    """The proportion of the cutoff distance at which the switch function starts."""
    graph_key: Optional[str] = "graph"
    """The key of the graph containing the distances and edge mask."""
    output_key: Optional[str] = None
    """The key of the output switch array. If None, it is added to the graph."""
    switch_type: str = "cosine"
    """The type of switch function. Can be 'cosine', 'polynomial', or 'exponential'."""
    p: Optional[float] = None
    """ The parameter of the switch function. If None, it is fixed to the default for each `switch_type`."""
    trainable: bool = False
    """Whether the switch parameter is trainable."""

    FID: ClassVar[str] = "SWITCH_FUNCTION"

    @nn.compact
    def __call__(self, inputs) -> Any:
        if self.graph_key is not None:
            graph = inputs[self.graph_key]
            distances, edge_mask = graph["distances"], graph["edge_mask"]
            if self.cutoff is not None:
                edge_mask = jnp.logical_and(edge_mask, (distances < self.cutoff))
                cutoff = self.cutoff
            else:
                cutoff = graph["cutoff"]
        else:
            # distances = inputs
            if len(inputs) == 3:
                distances, edge_mask, cutoff = inputs
            else:
                distances, edge_mask = inputs
                assert (
                    self.cutoff is not None
                ), "cutoff must be specified if no graph is given"
                # edge_mask = distances < self.cutoff
                cutoff = self.cutoff

        if self.switch_start > 1.0e-5:
            assert (
                self.switch_start < 1.0
            ), "switch_start is a proportion of cutoff and must be smaller than 1."
            cutoff_in = self.switch_start * cutoff
            x = distances - cutoff_in
            end = cutoff - cutoff_in
        else:
            x = distances
            end = cutoff

        switch_type = self.switch_type.lower()
        if switch_type == "cosine":
            p = self.p if self.p is not None else 1.0
            if self.trainable:
                p = self.param("p", lambda rng: jnp.asarray(p, dtype=jnp.float32))
            switch = (0.5 * jnp.cos(x * (jnp.pi / end)) + 0.5) ** p

        elif switch_type == "polynomial":
            p = self.p if self.p is not None else 3.0
            if self.trainable:
                p = self.param("p", lambda rng: jnp.asarray(p, dtype=jnp.float32))
            d = x / end
            switch = (
                1.0
                - 0.5 * (p + 1) * (p + 2) * d**p
                + p * (p + 2) * d ** (p + 1)
                - 0.5 * p * (p + 1) * d ** (p + 2)
            )

        elif switch_type == "exponential":
            p = self.p if self.p is not None else 1.0
            if self.trainable:
                p = self.param("p", lambda rng: jnp.asarray(p, dtype=jnp.float32))
            r2 = x**2
            c2 = end**2
            switch = jnp.exp(-p * r2 / (c2 - r2))
        
        elif switch_type == "hard":
            switch = jnp.where(distances < cutoff, 1.0, 0.0)
        else:
            raise ValueError(f"Unknown switch function {switch_type}")

        if self.switch_start > 1.0e-5:
            switch = jnp.where(distances < cutoff_in, 1.0, switch)

        switch = jnp.where(edge_mask, switch, 0.0)

        if self.graph_key is not None:
            if self.output_key is not None:
                return {**inputs, self.output_key: switch}
            else:
                return {**inputs, self.graph_key: {**graph, "switch": switch}}
        else:
            return switch  # , edge_mask



import sys, os, io
import argparse
from pathlib import Path

import numpy as np
import jax
import jax.numpy as jnp

from .models import FENNIX
from .utils.io import xyz_reader
from .utils.periodic_table import PERIODIC_TABLE_REV_IDX
from .utils.atomic_units import AtomicUnits as au
import yaml
import pickle
import json


def main():
    parser = argparse.ArgumentParser(prog="fennol_analyze")
    parser.add_argument(
        "input_file", type=Path, help="file containing the geometries to analyze"
    )
    parser.add_argument(
        "model_file", type=Path, help="file containing the model to use"
    )
    parser.add_argument("--output", "-o", type=Path, help="file to write the output to")
    parser.add_argument(
        "--batch_size",
        "-b",
        type=int,
        default=1,
        help="batch size to use for the model",
    )
    parser.add_argument(
        "--device", type=str, default="cpu", help="device to use for the model"
    )
    parser.add_argument("-f64", action="store_true", help="use double precision")
    parser.add_argument("--periodic", action="store_true", help="use PBC")
    parser.add_argument(
        "--format", type=str, default="xyz", help="format of the input file"
    )
    parser.add_argument(
        "--output_keys",
        type=str,
        nargs="+",
        help="keys to output",
        default=["total_energy"],
    )
    args = parser.parse_args()

    # set the device
    device = args.device.lower()
    if device == "cpu":
        device = "cpu"
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
    elif device.startswith("cuda") or device.startswith("gpu"):
        if ":" in device:
            num = device.split(":")[-1]
            os.environ["CUDA_VISIBLE_DEVICES"] = num
        else:
            os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        device = "gpu"

    _device = jax.devices(device)[0]
    jax.config.update("jax_default_device", _device)

    # set the precision
    if args.f64:
        jax.config.update("jax_enable_x64", True)
        fprec = "float64"
    else:
        fprec = "float32"
    jax.config.update("jax_default_matmul_precision", "highest")


    # check the input file
    input_file = args.input_file.resolve()
    assert input_file.exists(), f"Input file {input_file} does not exist"
    # assert args.format == "xyz", f"Only xyz format is supported for now"
    assert args.format in ["xyz","arc","pkl"], f"Only xyz, arc and pkl formats are supported for now"
    output_keys = args.output_keys
    xyz_indexed = args.format == "arc"

    # load the model
    model_file: Path = args.model_file.resolve()
    assert model_file.exists(), f"Model file {model_file} does not exist"
    model = FENNIX.load(model_file, use_atom_padding=True)

    # check the output file
    if args.output is not None:
        output_file: Path = args.output.resolve()
        assert not output_file.exists(), f"Output file {args.output} already exists"
        assert output_file.suffix in [
            ".pkl",
            ".json",
            ".yaml",
        ], f"Unsupported output file format {output_file.suffix}"

    # print metadata
    metadata = {
        "input_file": str(input_file),
        "model_file": str(model_file),
        "output_keys": output_keys,
        "energy_unit": model.energy_unit,
    }
    print("metadata:")
    for k,v in metadata.items():
        print(f"  {k}: {v}")

    # define the model prediction function
    def model_predict(batch):
        natoms = np.array([frame["natoms"] for frame in batch])
        batch_index = np.concatenate([frame["batch_index"] for frame in batch])
        species = np.concatenate([frame["species"] for frame in batch])
        xyz = np.concatenate([frame["coordinates"] for frame in batch], axis=0)
        inputs = {
            "species": species,
            "coordinates": xyz,
            "batch_index": batch_index,
            "natoms": natoms,}
        if args.periodic:
            cells = np.concatenate([frame["cell"] for frame in batch], axis=0)
            inputs["cells"] = cells

        if "forces" in output_keys:
            e, f, output = model.energy_and_forces(
                **inputs
            )
        else:
            e, output = model.total_energy(
                **inputs
            )
        return output

    # define the function to process a batch
    def process_batch(batch):
        output = model_predict(batch)
        natoms = np.array([frame["natoms"] for frame in batch])
        if args.periodic:
            cells = np.array(output["cells"])
        species = np.array(output["species"])
        coordinates = np.array(output["coordinates"])
        natshift = np.concatenate([np.array([0], dtype=np.int32), np.cumsum(natoms)])
        frames_data = []
        for i in range(len(batch)):
            frame_data = {
                "species": species[natshift[i] : natshift[i + 1]].tolist(),
                "coordinates": coordinates[natshift[i] : natshift[i + 1]].tolist(),
            }
            if args.periodic:
                frame_data["cell"] = cells[i].tolist()

            for k in output_keys:
                if k not in output:
                    raise ValueError(f"Output key {k} not found")
                v = output[k]
                if v.shape[0] == species.shape[0]:
                    frame_data[k] = v[natshift[i] : natshift[i + 1]].tolist()
                elif v.shape[0] == natoms.shape[0]:
                    frame_data[k] = v[i].tolist()
                else:
                    raise ValueError(f"Output key {k} has wrong shape")

            frames_data.append(frame_data)
        return frames_data

    ### start processing the input file
    # reader = xyz_reader(input_file, has_comment_line=True, indexed=xyz_indexed)
    if args.format in ["arc","xyz"]:
        def reader():
            for symbols, xyz, comment in xyz_reader(input_file, has_comment_line=True, indexed=xyz_indexed):
                species = np.array([PERIODIC_TABLE_REV_IDX[s] for s in symbols])
                inputs = {"species": species, "coordinates": xyz, "natoms":species.shape[0]}
                if args.periodic:
                    cell = np.array([float(x) for x in comment.split()]).reshape(1,3,3)
                    inputs["cell"] = cell
                yield inputs

    elif args.format == "pkl":
        with open(input_file, "rb") as f:
            data = pickle.load(f)
        if isinstance(data, dict):
            if "frames" in data:
                frames = data["frames"]
            elif "training" in data:
                frames = data["training"]
                if "validation" in data:
                    frames.extend(data["validation"])
            else:
                raise ValueError("No frames found in the input file")
        else:
            frames = data
        
        def reader():
            for frame in frames:
                species = np.array(frame["species"])
                coordinates = np.array(frame["coordinates"])
                inputs = {"species": species, "coordinates": coordinates,"natoms":species.shape[0]}
                if args.periodic:
                    cell = np.array(frame["cell"]).reshape(1,3,3)
                    inputs["cell"] = cell
                if "total_charge" in frame:
                    inputs["total_charge"] = frame["total_charge"]
                yield inputs

    batch = []
    output_data = []
    ibatch = 0
    for frame in reader():
        batch_index = np.full(frame["natoms"], ibatch, dtype=np.int32)
        frame["batch_index"] = batch_index
        batch.append(frame)
        ibatch += 1
        if len(batch) == args.batch_size:
            frames_data = process_batch(batch)
            if args.output is None:
                for frame_data in frames_data:
                    print("\n---\n")
                    for k in frame_data:
                        print(f"{k}: {frame_data[k]}")
            output_data.extend(frames_data)
            batch = []
            ibatch = 0
    # process the last batch
    if len(batch) > 0:
        frames_data = process_batch(batch)
        if args.output is None:
            for frame_data in frames_data:
                print("\n---\n")
                for k in frame_data:
                    print(f"{k}: {frame_data[k]}")
        output_data.extend(frames_data)

    # write the output to a file
    if args.output is not None:
        output_data = {
            "metadata": metadata,
            "frames": output_data,
        }
        if output_file.suffix == ".json":
            with open(args.output, "w") as f:
                json.dump(output_data, f, indent=2)
        elif output_file.suffix == ".yaml":
            with open(args.output, "w") as f:
                yaml.dump(output_data, f, sort_keys=False)
        elif output_file.suffix == ".pkl":
            with open(args.output, "wb") as f:
                pickle.dump(output_data, f)


if __name__ == "__main__":
    main()



import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, ClassVar
import numpy as np
import dataclasses

from ...utils.periodic_table import PERIODIC_TABLE, VALENCE_STRUCTURE
from ..misc.encodings import SpeciesEncoding, RadialBasis


class EEACSF(nn.Module):
    """element-embracing Atom-Centered Symmetry Functions 

    FID : EEACSF

    This is an embedding similar to ANI that include simple chemical information in the AEVs
    without trainable parameters (fixed embedding).
    The angle embedding is computed using a low-order Fourier expansion.

    ### Reference
    Loosely inspired from M. Eckhoff and M. Reiher, Lifelong Machine Learning Potentials,
    J. Chem. Theory Comput. 2023, 19, 12, 3509–3525, https://doi.org/10.1021/acs.jctc.3c00279

    """

    _graphs_properties: Dict
    graph_angle_key: str
    """ The key in the input dictionary that corresponds to the angular graph."""
    nmax_angle: int = 4
    """ The maximum fourier order for the angle representation."""
    embedding_key: str = "embedding"
    """ The key to use for the output embedding in the returned dictionary."""
    graph_key: str = "graph"
    """ The key in the input dictionary that corresponds to the radial graph."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """ The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """ The radial basis parameters the radial AEV. See `fennol.models.misc.encodings.RadialBasis`"""
    radial_basis_angle: dict = dataclasses.field(default_factory=dict)
    """ The radial basis parameters for the angular AEV. See `fennol.models.misc.encodings.RadialBasis`"""
    angle_combine_pairs: bool = False
    """ If True, the angular AEV is computed by combining pairs of radial AEV."""

    FID: ClassVar[str] = "EEACSF"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]

        # species encoding
        
        onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
            species
        )

        # Radial graph
        graph = inputs[self.graph_key]
        distances = graph["distances"]
        switch = graph["switch"][:,None]
        edge_src = graph["edge_src"]
        edge_dst = graph["edge_dst"]

        # Radial BASIS
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_terms = (
            RadialBasis(
                **{
                    **self.radial_basis,
                    "end": cutoff,
                    "name": f"RadialBasis",
                }
            )(distances)*switch
        )
        # aggregate radial AEV
        radial_aev = jax.ops.segment_sum(
            radial_terms[:, :, None] * onehot[edge_dst, None, :],
            edge_src,
            species.shape[0],
        ).reshape(species.shape[0], -1)

        # Angular graph
        graph_angle = inputs[self.graph_angle_key]     
        angles = graph_angle["angles"]
        dang = graph_angle["distances"]
        central_atom = graph_angle["central_atom"]
        angle_src, angle_dst = graph_angle["angle_src"], graph_angle["angle_dst"]
        switch_angles = graph_angle["switch"][:, None]
        angular_cutoff = self._graphs_properties[self.graph_angle_key]["cutoff"]
        edge_dst_ang = graph_angle["edge_dst"]


        radial_basis_angle = (
            self.radial_basis_angle
            if self.radial_basis_angle is not None
            else self.radial_basis
        )

        # Angular AEV parameters
        if self.angle_combine_pairs:
            factor2 = RadialBasis(
                **{
                    **radial_basis_angle,
                    "end": angular_cutoff,
                    "name": f"RadialBasisAng",
                }
            )(dang)*switch_angles
            radial_ang = (factor2[:, :, None] * onehot[edge_dst_ang, None, :]).reshape(-1, onehot.shape[1]*factor2.shape[1])

            radial_aev_ang = radial_ang[angle_src]*radial_ang[angle_dst]

            # Angular AEV
            nangles = jnp.asarray(
                np.arange(self.nmax_angle + 1)[None, :], dtype=angles.dtype
            )
            factor1 = jnp.cos(nangles * angles[:, None])

            angular_aev = jax.ops.segment_sum(
                factor1[:, None, :] * radial_aev_ang[:, :, None],
                central_atom,
                species.shape[0],
            ).reshape(species.shape[0], -1)

        else:
            valence = SpeciesEncoding(
                encoding="sjs_coordinates", name="SJSEncoding", trainable=False
            )(species)
            d12 = 0.5 * (dang[angle_src] + dang[angle_dst])
            switch12 = switch_angles[angle_src] * switch_angles[angle_dst]

            
            factor2 = RadialBasis(
                **{
                    **radial_basis_angle,
                    "end": angular_cutoff,
                    "name": f"RadialBasisAng",
                }
            )(d12)

            # Angular AEV
            nangles = jnp.asarray(
                np.arange(self.nmax_angle + 1)[None, :], dtype=angles.dtype
            )
            factor1 = jnp.cos(nangles * angles[:, None]) * switch12

            angular_terms = (factor1[:, None, :] * factor2[:, :, None]).reshape(
                -1, factor1.shape[1] * factor2.shape[1]
            )

            valence_dst = valence[edge_dst_ang]
            vangsrc = valence_dst[angle_src]
            vangdst = valence_dst[angle_dst]
            valence_ang_p = vangsrc + vangdst
            valence_ang_m = vangsrc * vangdst
            valence_ang = (valence_ang_p[:, :, None] * valence_ang_m[:, None, :]).reshape(
                -1, valence_ang_p.shape[1] * valence_ang_m.shape[1]
            )

            angular_aev = jax.ops.segment_sum(
                angular_terms[:, :, None] * valence_ang[:, None, :],
                central_atom,
                species.shape[0],
            ).reshape(species.shape[0], -1)

        
        embedding = jnp.concatenate((onehot, radial_aev, angular_aev), axis=-1)
        if self.embedding_key is None:
            return embedding
        return {**inputs, self.embedding_key: embedding}



import numpy as np
import jax.numpy as jnp
import jax
from functools import partial


def kernel_lorentz(w, w0, gamma):
    sel = np.logical_and(np.abs(w) < 1.0e-10, np.abs(w0) < 1.0e-10)
    w2 = np.where(sel, 1.0, w**2)
    w02 = np.where(sel, 1.0, w0**2)
    return gamma * w2 / (np.pi * (w2 * gamma**2 + (w2 - w02) ** 2))


def kernel_lorentz_pot(w, w0, gamma):
    sel = np.logical_and(jnp.abs(w) < 1.0e-10, np.abs(w0) < 1.0e-10)
    w2 = np.where(sel, 1.0, w**2)
    w02 = np.where(sel, 1.0, w0**2)
    return gamma * w2 / (np.pi * (w02 * gamma**2 + (w2 - w02) ** 2))


def deconvolute_spectrum(
    s_in,
    omega,
    gamma,
    niteration=10,
    kernel=kernel_lorentz,
    trans=False,
    symmetrize=True,
    thr=1.0e-10,
    verbose=False,
    K_D=None,
):
    assert s_in.shape[0] == omega.shape[0], "s_in and omega must have the same length"
    domega = omega[1] - omega[0]
    if symmetrize:
        nom_save = omega.shape[0]
        s_in = np.concatenate((np.flip(s_in[1:], axis=0), s_in), axis=0)
        omega = np.concatenate((-np.flip(omega[1:], axis=0), omega), axis=0)

    if K_D is not None:
        K, D = K_D
        assert K.shape == (omega.shape[0],omega.shape[0]), "K and omega must have the same length"
        assert D.shape == K.shape, "D and K must have the same shape"
    else:
        if verbose:
            print("deconvolution started.")
            print("  computing kernel matrix...")
        nom = omega.shape[0]
        omij0, omij1 = np.meshgrid(omega, omega)
        omij0 = omij0.flatten(order="F")
        omij1 = omij1.flatten(order="F")
        K = kernel(omij0, omij1, gamma).reshape(nom, nom)
        if trans:
            omnorm = np.arange(nom) * domega
            omnorm = np.concatenate((-np.flip(omnorm[1:], axis=0), omnorm), axis=0)
            omnormij0, omnormij1 = np.meshgrid(omega, omnorm)
            omnormij0 = omnormij0.flatten(order="F")
            omnormij1 = omnormij1.flatten(order="F")
            Knorm = kernel(omnormij0, omnormij1, gamma).reshape(nom, 2 * nom - 1)
            K = K / np.sum(Knorm, axis=1)[:, None] / domega
            del Knorm, omnormij0, omnormij1, omnorm
        else:
            K = K / np.sum(K, axis=0)[None, :] / domega

        if verbose:
            print("  kernel matrix computed.")
            print("  computing double convolution matrix...")

        D = (K.T @ K) * domega

        if verbose:
            print("  double convolution matrix computed.")
            print("  starting iterations.")

    s_out = s_in
    h = (K.T @ s_in) * domega
    for i in range(niteration):
        if verbose:
            print(f"  iteration {i+1}/{niteration}")
        den = (s_out[None, :] * D).sum(axis=1) * domega
        s_next = s_out * h / den
        diff = ((s_next - s_out) ** 2).sum() / (s_out**2).sum()
        if verbose:
           print(f"    relative difference: {diff}")
        s_out = s_next
        if diff < thr:
           break

    if verbose:
        print("deconvolution finished.")
    s_rec = (K @ s_out) * domega
    if symmetrize:
        s_rec = s_rec[nom_save - 1 :]
        s_out = s_out[nom_save - 1 :]

    return s_out, s_rec, (K, D)



import sys, os, io
import argparse
import time
from pathlib import Path

import numpy as np
from collections import defaultdict
import jax
import jax.numpy as jnp
import pickle

from ..utils.io import (
    write_arc_frame,
    write_xyz_frame,
    write_extxyz_frame,
    human_time_duration,
)
from .utils import wrapbox, save_dynamics_restart
from ..utils import minmaxone, AtomicUnits as au
from ..utils.input_parser import parse_input
from .integrate import initialize_dynamics


def main():
    # os.environ["OMP_NUM_THREADS"] = "1"
    sys.stdout = io.TextIOWrapper(
        open(sys.stdout.fileno(), "wb", 0), write_through=True
    )
    ### Read the parameter file
    parser = argparse.ArgumentParser(prog="fennol_md")
    parser.add_argument("param_file", type=Path, help="Parameter file")
    args = parser.parse_args()
    simulation_parameters = parse_input(args.param_file)

    ### Set the device
    device: str = simulation_parameters.get("device", "cpu").lower()
    if device == "cpu":
        device = "cpu"
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
    elif device.startswith("cuda") or device.startswith("gpu"):
        if ":" in device:
            num = device.split(":")[-1]
            os.environ["CUDA_VISIBLE_DEVICES"] = num
        else:
            os.environ["CUDA_VISIBLE_DEVICES"] = "0"
        device = "gpu"

    _device = jax.devices(device)[0]
    jax.config.update("jax_default_device", _device)

    ### Set the precision
    enable_x64 = simulation_parameters.get("double_precision", False)
    jax.config.update("jax_enable_x64", enable_x64)
    fprec = "float64" if enable_x64 else "float32"

    matmul_precision = simulation_parameters.get("matmul_prec", "highest").lower()
    assert matmul_precision in [
        "default",
        "high",
        "highest",
    ], "matmul_prec must be one of 'default','high','highest'"
    if matmul_precision != "highest":
        print(f"# Setting matmul precision to '{matmul_precision}'")
    if matmul_precision == "default" and fprec == "float32":
        print(
            "# Warning: default matmul precision involves float16 operations which may lead to large numerical errors on energy and pressure estimations ! It is recommended to set matmul_prec to 'high' or 'highest'."
        )
    jax.config.update("jax_default_matmul_precision", matmul_precision)

    # with jax.default_device(_device):
    dynamic(simulation_parameters, device, fprec)


def dynamic(simulation_parameters, device, fprec):
    tstart_dyn = time.time()

    random_seed = simulation_parameters.get(
        "random_seed", np.random.randint(0, 2**32 - 1)
    )
    print(f"# random_seed: {random_seed}")
    rng_key = jax.random.PRNGKey(random_seed)

    ## INITIALIZE INTEGRATOR AND SYSTEM
    rng_key, subkey = jax.random.split(rng_key)
    step, update_conformation, system_data, dyn_state, conformation, system = (
        initialize_dynamics(simulation_parameters, fprec, subkey)
    )

    nat = system_data["nat"]
    dt = dyn_state["dt"]
    ## get number of steps
    nsteps = int(simulation_parameters.get("nsteps"))
    start_time_ps = dyn_state.get("start_time_ps", 0.0)

    ### Set I/O parameters
    Tdump = simulation_parameters.get("tdump", 1.0 / au.PS) * au.FS
    ndump = int(Tdump / dt)
    system_name = system_data["name"]
    estimate_pressure = dyn_state["estimate_pressure"]

    @jax.jit
    def check_nan(system):
        return jnp.any(jnp.isnan(system["vel"])) | jnp.any(
            jnp.isnan(system["coordinates"])
        )

    if system_data["pbc"] is not None:
        cell = system["cell"]
        reciprocal_cell = np.linalg.inv(cell)
        do_wrap_box = simulation_parameters.get("wrap_box", False)
    else:
        cell = None
        reciprocal_cell = None
        do_wrap_box = False

    ### Energy units and print initial energy
    per_atom_energy = simulation_parameters.get("per_atom_energy", True)
    energy_unit_str = system_data["energy_unit_str"]
    energy_unit = system_data["energy_unit"]
    print("# Energy unit: ", energy_unit_str)
    atom_energy_unit = energy_unit
    atom_energy_unit_str = energy_unit_str
    if per_atom_energy:
        atom_energy_unit /= nat
        atom_energy_unit_str = f"{energy_unit_str}/atom"
        print("# Printing Energy per atom")
    print(
        f"# Initial potential energy: {system['epot']*atom_energy_unit}; kinetic energy: {system['ek']*atom_energy_unit}"
    )
    f = system["forces"]
    minmaxone(jnp.abs(f * energy_unit), "# forces min/max/rms:")

    ## printing options
    print_timings = simulation_parameters.get("print_timings", False)
    nprint = int(simulation_parameters.get("nprint", 10))
    assert nprint > 0, "nprint must be > 0"
    nsummary = simulation_parameters.get("nsummary", 100 * nprint)
    assert nsummary > nprint, "nsummary must be > nprint"

    save_keys = simulation_parameters.get("save_keys", [])
    if save_keys:
        print(f"# Saving keys: {save_keys}")
        fkeys = open(f"{system_name}.traj.pkl", "wb+")
    else:
        fkeys = None

    ### Print header
    include_thermostat_energy = "thermostat_energy" in system["thermostat"]
    thermostat_name = dyn_state["thermostat_name"]
    pimd = dyn_state["pimd"]
    variable_cell = dyn_state["variable_cell"]
    nbeads = system_data.get("nbeads", 1)
    dyn_name = "PIMD" if pimd else "MD"
    print("#" * 84)
    print(
        f"# Running {nsteps:_} steps of {thermostat_name} {dyn_name} simulation on {device}"
    )
    header = "#     Step   Time[ps]        Etot        Epot        Ekin    Temp[K]"
    if pimd:
        header += "  Temp_c[K]"
    if include_thermostat_energy:
        header += "      Etherm"
    if estimate_pressure:
        print_aniso_pressure = simulation_parameters.get("print_aniso_pressure", False)
        pressure_unit_str = simulation_parameters.get("pressure_unit", "atm")
        pressure_unit = au.get_multiplier(pressure_unit_str) * au.BOHR**3
        header += f"  Press[{pressure_unit_str}]"
    if variable_cell:
        header += "   Density"
    print(header)

    ### Open trajectory file
    traj_format = simulation_parameters.get("traj_format", "arc").lower()
    if traj_format == "xyz":
        traj_ext = ".traj.xyz"
        write_frame = write_xyz_frame
    elif traj_format == "extxyz":
        traj_ext = ".traj.extxyz"
        write_frame = write_extxyz_frame
    elif traj_format == "arc":
        traj_ext = ".arc"
        write_frame = write_arc_frame
    else:
        raise ValueError(
            f"Unknown trajectory format '{traj_format}'. Supported formats are 'arc' and 'xyz'"
        )

    write_all_beads = simulation_parameters.get("write_all_beads", False) and pimd

    if write_all_beads:
        fout = [
            open(f"{system_name}_bead{i+1:03d}" + traj_ext, "a") for i in range(nbeads)
        ]
    else:
        fout = open(system_name + traj_ext, "a")

    ensemble_key = simulation_parameters.get("etot_ensemble_key", None)
    if ensemble_key is not None:
        fens = open(f"{system_name}.ensemble_weights.traj", "a")

    write_centroid = simulation_parameters.get("write_centroid", False) and pimd
    if write_centroid:
        fcentroid = open(f"{system_name}_centroid" + traj_ext, "a")

    fcolvars = None

    ### initialize proprerty trajectories
    properties_traj = defaultdict(list)

    ### initialize counters and timers
    t0 = time.time()
    t0dump = t0
    istep = 0
    t0full = time.time()
    force_preprocess = False

    for istep in range(1, nsteps + 1):

        ### update the system
        dyn_state, system, conformation, model_output = step(
            istep, dyn_state, system, conformation, force_preprocess
        )

        ### print properties
        if istep % nprint == 0:
            t1 = time.time()
            tperstep = (t1 - t0) / nprint
            t0 = t1
            nsperday = (24 * 60 * 60 / tperstep) * dt / 1e6

            ek = system["ek"]
            epot = system["epot"]
            etot = ek + epot
            temper = 2 * ek / (3.0 * nat) * au.KELVIN

            th_state = system["thermostat"]
            if include_thermostat_energy:
                etherm = th_state["thermostat_energy"]
                etot = etot + etherm

            properties_traj[f"Etot[{atom_energy_unit_str}]"].append(
                etot * atom_energy_unit
            )
            properties_traj[f"Epot[{atom_energy_unit_str}]"].append(
                epot * atom_energy_unit
            )
            properties_traj[f"Ekin[{atom_energy_unit_str}]"].append(
                ek * atom_energy_unit
            )
            properties_traj["Temper[Kelvin]"].append(temper)
            if pimd:
                ek_c = system["ek_c"]
                temper_c = 2 * ek_c / (3.0 * nat) * au.KELVIN
                properties_traj["Temper_c[Kelvin]"].append(temper_c)

            ### construct line of properties
            simulated_time = start_time_ps + istep * dt / 1000
            line = f"{istep:10.6g} {simulated_time: 10.3f}  {etot*atom_energy_unit: #10.4f}  {epot*atom_energy_unit: #10.4f}  {ek*atom_energy_unit: #10.4f} {temper: 10.2f}"
            if pimd:
                line += f" {temper_c: 10.2f}"
            if include_thermostat_energy:
                line += f"  {etherm*atom_energy_unit: #10.4f}"
                properties_traj[f"Etherm[{atom_energy_unit_str}]"].append(
                    etherm * atom_energy_unit
                )
            if estimate_pressure:
                pres = system["pressure"] * pressure_unit
                properties_traj[f"Pressure[{pressure_unit_str}]"].append(pres)
                if print_aniso_pressure:
                    pres_tensor = system["pressure_tensor"] * pressure_unit
                    pres_tensor = 0.5 * (pres_tensor + pres_tensor.T)
                    properties_traj[f"Pressure_xx[{pressure_unit_str}]"].append(
                        pres_tensor[0, 0]
                    )
                    properties_traj[f"Pressure_yy[{pressure_unit_str}]"].append(
                        pres_tensor[1, 1]
                    )
                    properties_traj[f"Pressure_zz[{pressure_unit_str}]"].append(
                        pres_tensor[2, 2]
                    )
                    properties_traj[f"Pressure_xy[{pressure_unit_str}]"].append(
                        pres_tensor[0, 1]
                    )
                    properties_traj[f"Pressure_xz[{pressure_unit_str}]"].append(
                        pres_tensor[0, 2]
                    )
                    properties_traj[f"Pressure_yz[{pressure_unit_str}]"].append(
                        pres_tensor[1, 2]
                    )
                line += f" {pres:10.3f}"
            if variable_cell:
                density = system["density"]
                properties_traj["Density[g/cm^3]"].append(density)
                if print_aniso_pressure:
                    cell = system["cell"]
                    properties_traj[f"Cell_Ax[Angstrom]"].append(cell[0, 0])
                    properties_traj[f"Cell_Ay[Angstrom]"].append(cell[0, 1])
                    properties_traj[f"Cell_Az[Angstrom]"].append(cell[0, 2])
                    properties_traj[f"Cell_Bx[Angstrom]"].append(cell[1, 0])
                    properties_traj[f"Cell_By[Angstrom]"].append(cell[1, 1])
                    properties_traj[f"Cell_Bz[Angstrom]"].append(cell[1, 2])
                    properties_traj[f"Cell_Cx[Angstrom]"].append(cell[2, 0])
                    properties_traj[f"Cell_Cy[Angstrom]"].append(cell[2, 1])
                    properties_traj[f"Cell_Cz[Angstrom]"].append(cell[2, 2])
                line += f" {density:10.4f}"
                if "piston_temperature" in system["barostat"]:
                    piston_temperature = system["barostat"]["piston_temperature"]
                    properties_traj["T_piston[Kelvin]"].append(piston_temperature)

            print(line)

            ### save colvars
            if "colvars" in system:
                colvars = system["colvars"]
                if fcolvars is None:
                    # open colvars file and print header
                    fcolvars = open(f"{system_name}.colvars", "a")
                    fcolvars.write("# time[ps]")
                    fcolvars.write(" ".join(colvars.keys()))
                    fcolvars.write("\n")
                fcolvars.write(f"{simulated_time:.3f} ")
                fcolvars.write(" ".join([f"{v:.6f}" for v in colvars.values()]))
                fcolvars.write("\n")
                fcolvars.flush()

            if save_keys:
                data = {
                    k: (
                        np.asarray(model_output[k])
                        if isinstance(model_output[k], jnp.ndarray)
                        else model_output[k]
                    )
                    for k in save_keys
                }
                data["properties"] = {
                    k: float(v) for k, v in zip(header.split()[1:], line.split())
                }
                pickle.dump(data, fkeys)

        ### save frame
        if istep % ndump == 0:
            line = "# Write XYZ frame"
            if variable_cell:
                cell = np.array(system["cell"])
                reciprocal_cell = np.linalg.inv(cell)
            if do_wrap_box:
                if pimd:
                    centroid = wrapbox(system["coordinates"][0], cell, reciprocal_cell)
                    system["coordinates"] = system["coordinates"].at[0].set(centroid)
                else:
                    system["coordinates"] = wrapbox(
                        system["coordinates"], cell, reciprocal_cell
                    )
                conformation = update_conformation(conformation, system)
                line += " (atoms have been wrapped into the box)"
                force_preprocess = True
            print(line)

            save_dynamics_restart(system_data, conformation, dyn_state, system)

            properties = {
                "energy": float(system["epot"]) * energy_unit,
                "Time_ps": start_time_ps + istep * dt / 1000,
                "energy_unit": energy_unit_str,
            }

            if write_all_beads:
                coords = np.asarray(conformation["coordinates"].reshape(-1, nat, 3))
                for i, fb in enumerate(fout):
                    write_frame(
                        fb,
                        system_data["symbols"],
                        coords[i],
                        cell=cell,
                        properties=properties,
                        forces=None,  # np.asarray(system["forces"].reshape(nbeads, nat, 3)[0]) * energy_unit,
                    )
            else:
                write_frame(
                    fout,
                    system_data["symbols"],
                    np.asarray(conformation["coordinates"].reshape(-1, nat, 3)[0]),
                    cell=cell,
                    properties=properties,
                    forces=None,  # np.asarray(system["forces"].reshape(nbeads, nat, 3)[0]) * energy_unit,
                )
            if write_centroid:
                centroid = np.asarray(system["coordinates"][0])
                write_frame(
                    fcentroid,
                    system_data["symbols"],
                    centroid,
                    cell=cell,
                    properties=properties,
                    forces=np.asarray(system["forces"].reshape(nbeads, nat, 3)[0])
                    * energy_unit,
                )
            if ensemble_key is not None:
                weights = " ".join(
                    [f"{w:.6f}" for w in system["ensemble_weights"].tolist()]
                )
                fens.write(f"{weights}\n")
                fens.flush()

        ### summary over last nsummary steps
        if istep % (nsummary) == 0:
            if check_nan(system):
                raise ValueError(f"dynamics crashed at step {istep}.")
            tfull = time.time() - t0full
            t0full = time.time()
            tperstep = tfull / (nsummary)
            nsperday = (24 * 60 * 60 / tperstep) * dt / 1e6
            elapsed_time = time.time() - tstart_dyn
            estimated_remaining_time = tperstep * (nsteps - istep)
            estimated_total_time = elapsed_time + estimated_remaining_time

            print("#" * 50)
            print(f"# Step {istep:_} of {nsteps:_}  ({istep/nsteps*100:.5g} %)")
            print(f"# Simulated time      : {istep * dt*1.e-3:.3f} ps")
            print(f"# Tot. Simu. time     : {start_time_ps + istep * dt*1.e-3:.3f} ps")
            print(f"# Tot. elapsed time   : {human_time_duration(elapsed_time)}")
            print(
                f"# Est. total duration   : {human_time_duration(estimated_total_time)}"
            )
            print(
                f"# Est. remaining time : {human_time_duration(estimated_remaining_time)}"
            )
            print(f"# Time for {nsummary:_} steps : {human_time_duration(tfull)}")

            corr_kin = system["thermostat"].get("corr_kin", None)
            if corr_kin is not None:
                print(f"# QTB kin. correction : {100*(corr_kin-1.):.2f} %")
            print(f"# Averages over last {nsummary:_} steps :")
            for k, v in properties_traj.items():
                if len(properties_traj[k]) == 0:
                    continue
                mu = np.mean(properties_traj[k])
                sig = np.std(properties_traj[k])
                ksplit = k.split("[")
                name = ksplit[0].strip()
                unit = ksplit[1].replace("]", "").strip() if len(ksplit) > 1 else ""
                print(f"#   {name:10} : {mu: #10.5g}   +/- {sig: #9.3g}  {unit}")

            print(f"# Perf.: {nsperday:.2f} ns/day  ( {1.0 / tperstep:.2f} step/s )")
            print("#" * 50)
            if istep < nsteps:
                print(header)
            ## reset property trajectories
            properties_traj = defaultdict(list)

    print(f"# Run done in {human_time_duration(time.time()-tstart_dyn)}")
    ### close trajectory file
    fout.close()
    if ensemble_key is not None:
        fens.close()
    if fcolvars is not None:
        fcolvars.close()
    if fkeys is not None:
        fkeys.close()
    if write_centroid:
        fcentroid.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import sys
import time
import numpy as np


def xyz_reader(
    filename,
    box_info=False,
    indexed=False,
    start=1,
    stop=-1,
    step=1,
    max_frames=None,
    stream=False,
    interval=2.0,
    sleep=time.sleep,
):
    if stop > 0 and start > stop:
        return
    pf = open(filename, "r")
    inside_frame = False
    nat_read = False
    iframe = 0
    stride = step
    nframes = 0
    if start > 1:
        print(f"skipping {start-1} frames")
    if not box_info:
        box = None
    while True:
        line = pf.readline()

        if not line:
            if stream:
                sleep(interval)
                continue
            elif inside_frame or nat_read:
                raise Exception("Error: premature end of file!")
            else:
                break

        line = line.strip()
        if line.startswith("#"):
            continue

        if not inside_frame:
            if not nat_read:
                nat = int(line.split()[0])
                nat_read = True
                iat = 0
                inside_frame = not box_info
                xyz = np.zeros((nat, 3))
                species = []
                continue

            box = np.array(line.split(), dtype="float")
            inside_frame = True
            continue

        ls = line.split()
        iat, s = (int(ls[0]), 1) if indexed else (iat + 1, 0)
        species.append(ls[s])
        xyz[iat - 1, :] = np.array([ls[s + 1], ls[s + 2], ls[s + 3]], dtype="float")
        if iat == nat:
            iframe += 1
            inside_frame = False
            nat_read = False
            if iframe < start:
                continue
            stride += 1
            if stride >= step:
                nframes += 1
                stride = 0
                yield np.array(species, dtype="<U2"), xyz, box
            if (max_frames is not None and nframes >= max_frames) or (
                stop > 0 and iframe >= stop
            ):
                break


def apply_pbc(xyz, pairs, cell, rmax, minimum_image):
    cellinv = np.linalg.inv(cell)
    if minimum_image:
        vec = xyz[pairs[1]] - xyz[pairs[0]]
        # q = cellinv @ vec.T
        shift = -np.round(
            np.dot(vec, cellinv.T)
        )  # np.einsum("ij,sj->si", cellinv, vec)
        vec = vec + np.dot(shift, cell.T)  # np.einsum("ij,sj->si", cell, q)
        # return (cell @ q).T
    else:
        shift = -np.floor(np.dot(xyz, cellinv.T))
        xyz_pbc = xyz + np.dot(shift, cell.T)

        inv_distances = np.sum(cellinv**2, axis=0) ** 0.5
        num_repeats = np.ceil(rmax * inv_distances).astype(np.int32)
        cell_shift_pbc = np.array(
            np.meshgrid(*[np.arange(-n, n + 1) for n in num_repeats])
        ).T.reshape(-1, 3)
        dvec = np.dot(cell_shift_pbc, cell.T)

        vec = (
            (xyz_pbc[pairs[1]] - xyz_pbc[pairs[0]])[:, None, :] + dvec[None, :, :]
        ).reshape(-1, 3)
    return vec


def get_pairs(species, sp1=None, sp2=None):
    # nat=len(species)
    # id1=np.flatnonzero(species==sp1) if sp1 is not None else np.arange(nat)
    # if sp2 != sp1:
    #  id2=np.flatnonzero(species==sp2) if sp2 is not None else np.arange(nat)
    # pairs=np.array([np.repeat(id1, len(id2)),np.tile(id2, len(id1))])
    # if sp1 is None or sp2 is None or sp1==sp2:
    #  pairs=pairs[:,pairs[0]!=pairs[1]]
    # return pairs
    # if sp1 is None and sp2 is not None: sp1=sp2
    # if sp2 is None and sp1 is not None: sp2=sp1
    if sp1 is None:
        if sp2 is None:
            return np.array(np.triu_indices(len(species), 1)), 0.5 * len(species) ** 2
        sp1 = sp2
    id1 = np.flatnonzero(species == sp1)
    num1 = len(id1)
    if sp2 == sp1 or sp2 is None:
        pairsloc = np.triu_indices(len(id1), 1)
        return np.array([id1[pairsloc[0]], id1[pairsloc[1]]]), 0.5 * num1**2
    id2 = np.flatnonzero(species == sp2)
    num2 = len(id2)
    return np.array([np.repeat(id1, len(id2)), np.tile(id2, len(id1))]), num1 * num2


def radial_density(
    species,
    xyz,
    cell=None,
    rmax=10.0,
    dr=0.1,
    sp1=None,
    sp2=None,
    pairs=None,
    timer=False,
    minimum_image=False,
):
    nat = len(species)
    if timer:
        print(sp1, sp2)
        time0 = time.time()
    if pairs is None:
        pairs, n_pairs_norm = get_pairs(species, sp1, sp2)
    n_pairs = pairs.shape[1]
    mult = 1 if sp1 != sp2 else 2
    # print(mult*n_pairs,mult)
    if timer:
        print("pairs:", time.time() - time0)
        time0 = time.time()
    if cell is not None:
        vec = apply_pbc(xyz, pairs, cell, rmax, minimum_image)
    else:
        vec = xyz[pairs[0]] - xyz[pairs[1]]
    dist = np.linalg.norm(vec, axis=-1)
    if timer:
        print("dists:", time.time() - time0)
        time0 = time.time()
    n_bins = int(rmax / dr)
    bins = np.linspace(0, rmax, n_bins + 1)
    r, _ = np.histogram(dist, bins=n_bins, range=(0, rmax))
    if timer:
        print("hist:", time.time() - time0)
    expect = n_pairs_norm * 4.0 / 3.0 * np.pi * ((bins[1:]) ** 3 - bins[:-1] ** 3)
    if cell is not None:
        volume = np.dot(cell[:, 0], np.cross(cell[:, 1], cell[:, 2]))
        expect /= volume
    return r / expect


def radial_from_file(
    xyzfile,
    pairs,
    outfile="gr.dat",
    rmax=10,
    dr=0.02,
    thermalize=0,
    stride=1,
    box_info=False,
    indexed=False,
    watch=False,
    cell=None,
    minimum_image=False,
    **kwargs,
):
    if "tinker" in kwargs and kwargs["tinker"]:
        indexed = True
        box_info = True

    try:
        # frames = list(read_arc_file(arcfile,max_frames=205))
        # print(len(frames))
        reader = xyz_reader(
            xyzfile,
            box_info=box_info,
            indexed=indexed,
            start=thermalize + 1,
            step=stride,
            stream=watch,
        )
        n_bins = int(rmax / dr)
        centers = (np.arange(n_bins) + 0.5) * dr
        nframe = 0
        if cell is not None:
            # cell=np.array([float(c) for c in cell.split()]).reshape((3,3),order="F")
            cell = np.array(cell).reshape((3, 3), order="F")
            print("cell matrix=")
            print(cell)
        # setup_figures(["g(r)"])
        # plt.pause(1.)
        pairs = [p.strip().replace("-", " ").replace("_", " ") for p in pairs]
        header = " ".join(["r"] + ["-".join(p.split()) for p in pairs])
        box = None
        gr_avg = np.zeros((n_bins, len(pairs)))
        for frame in reader:
            species, xyz, box = frame
            if box is not None:
                cell = np.diag(box[:3])
            nframe += 1
            print(nframe)
            for i, pair in enumerate(pairs):
                sp1, sp2 = pair.split()
                gr = radial_density(
                    species, xyz, cell, rmax, dr, sp1=sp1, sp2=sp2, timer=False, minimum_image=minimum_image
                )
                # gr=radial_density_numba(species,xyz,box,rmax,dr,sp1=sp1,sp2=sp2)
                gr_avg[:, i] += gr
            # cleanup_ax(plt.gca())
            # plt.plot(centers,gr_avg/nframe)
            # plt.pause(1.)
            np.savetxt(
                outfile, np.column_stack((centers, gr_avg / nframe)), header=header
            )

    except KeyboardInterrupt:
        sys.exit(0)


def main(xyzfile: str, thermalize: int = 0, watch: bool = False, minimum_image=False):

    # box length
    a = 9.3215
    cell = np.array([a, 0, 0, 0, a, 0, 0, 0, a]).reshape((3, 3), order="F")

    # maximum distance to compute the rdf
    rmax = 9.
    # bin width
    dr = 0.05

    # pairs of atoms to compute the rdf
    pairs = ["O O", "O H", "H H"]

    radial_from_file(
        xyzfile,
        pairs,
        cell=cell,
        rmax=rmax,
        dr=dr,
        indexed=True,
        box_info=False,
        watch=watch,
        thermalize=thermalize,
        minimum_image=minimum_image,
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="compute radial distribution function")
    parser.add_argument("arcfile", type=str, help="arc file")
    parser.add_argument(
        "--thermalize", "-t", type=int, default=0, help="number of frames to skip"
    )
    parser.add_argument("--watch", "-w", action="store_true", help="watch file")
    parser.add_argument("--minimage", action="store_true", help="watch file")
    args = parser.parse_args()
    main(args.arcfile, args.thermalize, args.watch, args.minimage)



import jax
import jax.numpy as jnp
import flax.linen as nn
import numpy as np
from typing import Any, Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils import AtomicUnits as au
import dataclasses
from ...utils.periodic_table import (
    D3_ELECTRONEGATIVITIES,
    D3_HARDNESSES,
    D3_VDW_RADII,
    D3_COV_RADII,
    D3_KAPPA,
    VDW_RADII,
    VALENCE_ELECTRONS,
    PAULING_ELECTRONEGATIVITY,
)


class CND4(nn.Module):
    """ Coordination number as defined in D4 dispersion correction
    
    FID : CN_D4
    """
    graph_key: str = "graph"
    """ The key for the graph input."""
    output_key: Optional[str] = None
    """ The key for the output."""
    k0: float = 7.5
    k1: float = 4.1
    k2: float = 19.09
    k3: float = 254.56
    electronegativity_factor: bool = False
    """ Whether to include electronegativity factor."""
    trainable: bool = False
    """ Whether the parameters are trainable."""

    FID: ClassVar[str]  = "CN_D4"

    @nn.compact
    def __call__(self, inputs):
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        species = inputs["species"]

        if self.trainable:
            rc = self.param("rc", lambda key: jnp.asarray(D3_COV_RADII))[species]
        else:
            rc = jnp.asarray(D3_COV_RADII)[species]
        rcij = rc[edge_src] + rc[edge_dst]
        rij = graph["distances"] / au.BOHR

        if self.trainable:
            k0 = self.k0 * jnp.abs(self.param("k0", lambda key: jnp.asarray(1.0)))
        else:
            k0 = self.k0

        CNij = (
            0.5 * (1 + jax.scipy.special.erf(-k0 * (rij / rcij - 1.))) * graph["switch"]
        )

        if self.electronegativity_factor:
            k1 = self.k1
            k2 = self.k2
            k3 = self.k3
            if self.trainable:
                k1 = k1 * jnp.abs(self.param("k1", lambda key: jnp.asarray(1.0)))
                k2 = self.param("k2", lambda key: jnp.asarray(1.0))
                k3 = jnp.abs(self.param("k3", lambda key: jnp.asarray(1.0)))
                en = self.param("en", lambda key: jnp.asarray(D3_ELECTRONEGATIVITIES))[
                    species
                ]
            else:
                en = jnp.asarray(PAULING_ELECTRONEGATIVITY)[species]
            en_ij = jnp.abs(en[edge_src] - en[edge_dst])
            dij = k1 * jnp.exp(-((en_ij + k2) ** 2) / k3)
            CNij = CNij * dij
        CNi = jax.ops.segment_sum(CNij, edge_src, species.shape[0])

        output_key = self.name if self.output_key is None else self.output_key
        return {**inputs, output_key: CNi, output_key + "_pair": CNij}


class SumSwitch(nn.Module):
    """Sum (a power of) the switch values for each neighbor.
    
    FID : SUM_SWITCH

    """
    graph_key: str = "graph"
    """ The key for the graph input."""
    output_key: Optional[str] = None
    """ The key for the output."""
    pow: float = 1.0
    """ The power to raise the switch values to."""
    trainable: bool = False
    """ Whether the pow parameter is trainable."""

    FID: ClassVar[str]  = "SUM_SWITCH"

    @nn.compact
    def __call__(self, inputs):
        graph = inputs[self.graph_key]
        edge_src = graph["edge_src"]
        switch = graph["switch"]

        if self.trainable:
            p = jnp.abs(
                self.param("pow", lambda key: jnp.asarray(self.pow))
            )
        else:
            p = self.pow
        shift=(1.e-3)**p

        cn = jax.ops.segment_sum((1.e-3+switch)**p-shift, edge_src, inputs["species"].shape[0])

        output_key = self.name if self.output_key is None else self.output_key
        return {**inputs, output_key: cn}


class CNShift(nn.Module):
    
    cn_key: str
    output_key: Optional[str] = None
    kappa_key: Optional[str] = None
    sqrt_shift: float = 1.0e-6
    ref_value: Union[str, float] = 1.0
    enforce_positive: bool = False
    cn_pow: float = 0.5

    FID: ClassVar[str]  = "CN_SHIFT"


    @nn.compact
    def __call__(self, inputs):
        CNi = inputs[self.cn_key]
        if self.kappa_key is not None:
            kappai = inputs[self.kappa_key]
            assert kappai.shape == CNi.shape
        else:
            species = inputs["species"]
            kappai = self.param("kappa", nn.initializers.zeros, (len(D3_COV_RADII),))[
                species
            ]
        shift = kappai * (CNi + self.sqrt_shift) ** self.cn_pow

        if isinstance(self.ref_value, str):
            ref_value = inputs[self.ref_value]
            assert ref_value.shape == shift.shape
        else:
            ref_value = self.ref_value

        if self.enforce_positive:
            shift = jax.nn.celu(shift, alpha=ref_value)

        out = ref_value + shift

        output_key = self.name if self.output_key is None else self.output_key
        return {**inputs, output_key: out}


class CNStore(nn.Module):
    cn_key: str
    output_key: Optional[str] = None
    store_size: int = 10
    n_gaussians: int = 4
    isolated_value: float = 0.0
    init_scale_cn: float = 5.0
    init_scale_values: float = 1.0
    beta: float = 6.0
    trainable: bool = True
    output_dim: int = 1
    squeeze: bool = True

    FID: ClassVar[str]  = "CN_STORE"

    @nn.compact
    def __call__(self, inputs):
        cn = inputs[self.cn_key]
        species = inputs["species"]

        cn_refs = self.param(
            "cn_refs",
            nn.initializers.uniform(self.init_scale_cn),
            (len(D3_COV_RADII), self.store_size),
        )[species]
        values_refs = self.param(
            "values_refs",
            nn.initializers.uniform(self.init_scale_values),
            (len(D3_COV_RADII), self.store_size, self.output_dim),
        )[species]

        beta = self.beta
        if self.trainable:
            beta = self.param("beta", lambda key: jnp.asarray(self.beta))
        j = jnp.asarray(np.arange(self.n_gaussians)[None, None, :], dtype=jnp.float32)
        delta_cns = jnp.log(
            jnp.sum(
                jnp.exp(-beta * j * ((cn[:, None] - cn_refs) ** 2)[:, :, None]), axis=-1
            )
        )
        w = jax.nn.softmax(delta_cns, axis=-1)

        values = jnp.sum(w[:, :, None] * values_refs, axis=1)
        if self.output_dim == 1 and self.squeeze:
            values = jnp.squeeze(values, axis=-1)

        output_key = self.name if self.output_key is None else self.output_key
        return {**inputs, output_key: values}


class FlatBottom(nn.Module):
    """Flat bottom potential energy surface.
    
    Realized by Côme Cattin, 2024.

    Flat bottom potential energy:
    E = alpha * (r - req) ** 2 if r >= req
    E = 0 if r < req

    FID: FLAT_BOTTOM
    """

    energy_key: Optional[str] = None
    """Key of the energy in the outputs."""
    graph_key: str = "graph"
    """Key of the graph in the inputs."""
    alpha: float = 400.0
    """Force constant of the flat bottom potential (in kcal/mol/A^2)."""
    r_eq_factor: float = 1.3
    """Factor to multiply the sum of the VDW radii of the two atoms."""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str] = "FLAT_BOTTOM"

    @nn.compact
    def __call__(self, inputs):

        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        distances = graph["distances"]
        rij = distances / au.BOHR
        training = "training" in inputs.get("flags", {})

        output = {}
        energy_key = self.energy_key if self.energy_key is not None else self.name

        if training:
            output[energy_key] =  jnp.zeros(species.shape[0],dtype=distances.dtype)
            return {**inputs, **output}

        # req is the sum of the covalent radii of the two atoms
        rcov = jnp.asarray(D3_COV_RADII)[species]
        req = self.r_eq_factor * (rcov[edge_src] + rcov[edge_dst])

        alpha = inputs.get("alpha", self.alpha)/ au.KCALPERMOL*au.BOHR**2

        flat_bottom_energy = jnp.where(
            rij > req, alpha  * (rij - req) ** 2, 0.
        )

        flat_bottom_energy = jax.ops.segment_sum(flat_bottom_energy, edge_src, num_segments=species.shape[0])

        energy_unit = au.get_multiplier(self._energy_unit)
        output[energy_key] = flat_bottom_energy * energy_unit

        return {**inputs, **output}


import argparse
from pathlib import Path
import yaml
import json
import os
from flax import traverse_util
import jax
import dataclasses

from .fennix import FENNIX


class IndentDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)


def main():
    # os.environ["CUDA_VISIBLE_DEVICES"] = ""
    _device = jax.devices("cpu")[0]
    jax.config.update("jax_default_device", _device)
    ### Read the parameter file
    parser = argparse.ArgumentParser(prog="fennol_inspect")
    parser.add_argument("model_file", type=Path, help="Model file")
    parser.add_argument(
        "-s",
        "--short",
        action="store_true",
        help="Print only module names in the list of modules",
    )
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="Print all module attributes and automatically added modules",
    )
    parser.add_argument(
        "-p","--prm", action="store_true", help="Print parameter shapes"
    )
    args = parser.parse_args()
    model_file = args.model_file
    # param_shapes = args.param_shapes
    # short = args.short
    # all = args.all

    ### Load the model
    if model_file.suffix == ".yaml" or model_file.suffix == ".yml":
        with open(model_file, "r") as f:
            model_dict = yaml.safe_load(f)
        model = FENNIX(**model_dict,rng_key=jax.random.PRNGKey(0))
    else:
        model = FENNIX.load(model_file)
    print(f"# filename: {model_file}")
    print(inspect_model(model, **vars(args)))


def inspect_model(model, prm=False, short=False, all=False, **kwargs):

    model_dict = model._input_args

    inspect_dict = {
        "energy_unit": model.energy_unit,
        "energy_terms": model.energy_terms,
        "cutoff": model.cutoff,
    }
    
    inspect_dict["preprocessing"] = dict(model_dict["preprocessing"])
    if not all:
        inspect_dict["modules"] = dict(model_dict["modules"])
    else:
        mods = {}
        for mod, inp in model.modules.layers:
            m = mod(**inp)
            key = m.name
            if key is None or key.startswith("_"):
                continue
            mods[key] = {}
            for field in dataclasses.fields(m):
                if field.name not in ["name", "parent"]:
                    mods[key][field.name] = getattr(m, field.name)
        inspect_dict["modules"] = mods
    
    if short:
        inspect_dict["preprocessing"] = list(inspect_dict["preprocessing"].keys())
        inspect_dict["modules"] = list(inspect_dict["modules"].keys())
    
    data = "# MODEL DESCRIPTION\n"
    data += yaml.dump(inspect_dict, sort_keys=False, Dumper=IndentDumper)

    params = model.variables["params"] if "params" in model.variables else {}

    if prm:
        shapes = traverse_util.path_aware_map(
            lambda p, v: f"[{','.join(str(i) for i in v.shape)}]",
            params,
        )

        data = (
            data
            + "\n\n# PARAMETER SHAPES\n"
            + yaml.dump(
                shapes,
                sort_keys=False,
            )
        )

    number_of_params = sum(
        jax.tree.leaves(
            traverse_util.path_aware_map(
                lambda p, v: v.size,
                params,
            )
        )
    )

    data += f"\n# NUMBER OF PARAMETERS: {number_of_params:_}"
    return data


if __name__ == "__main__":
    main()
